@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2014 Fanout, Inc.
+ * Copyright (C) 2014-2016 Fanout, Inc.
  *
  * This file is part of Pushpin.
  *
@@ -23,6 +23,7 @@
 #include <QByteArray>
 #include <QList>
 #include <QVariant>
+#include <QUrl>
 
 class WsControlPacket
 {
@@ -36,21 +37,30 @@ public:
 			KeepAlive,
 			Gone,
 			Grip,
+			NeedKeepAlive,
 			Cancel,
 			Send,
+			KeepAliveSetup,
+			Close,
 			Detach
 		};
 
 		QByteArray cid;
 		Type type;
+		QUrl uri;
 		QByteArray contentType;
 		QByteArray message;
+		int code;
+		QByteArray route;
 		QByteArray channelPrefix;
 		int ttl;
+		int timeout;
 
 		Item() :
 			type((Type)-1),
-			ttl(-1)
+			code(-1),
+			ttl(-1),
+			timeout(-1)
 		{
 		}
 	};
