@@ -73,7 +73,7 @@ public:
 	int peerCloseCode;
 	QVariant userData;
 	bool pendingUpdate;
-	WebSocket::ErrorCondition errorCondition;
+	ErrorCondition errorCondition;
 	QTimer *expireTimer;
 	QTimer *keepAliveTimer;
 	QList<Frame> inFrames;
@@ -108,11 +108,11 @@ public:
 		outContentType((int)Frame::Text)
 	{
 		expireTimer = new QTimer(this);
-		connect(expireTimer, SIGNAL(timeout()), this, SLOT(expire_timeout()));
+		connect(expireTimer, &QTimer::timeout, this, &Private::expire_timeout);
 		expireTimer->setSingleShot(true);
 
 		keepAliveTimer = new QTimer(this);
-		connect(keepAliveTimer, SIGNAL(timeout()), this, SLOT(keepAlive_timeout()));
+		connect(keepAliveTimer, &QTimer::timeout, this, &Private::keepAlive_timeout);
 	}
 
 	~Private()
@@ -1010,6 +1010,11 @@ void ZWebSocket::setIgnoreTlsErrors(bool on)
 	d->ignoreTlsErrors = on;
 }
 
+void ZWebSocket::setIsTls(bool on)
+{
+	d->requestUri.setScheme(on ? "wss" : "ws");
+}
+
 void ZWebSocket::start(const QUrl &uri, const HttpHeaders &headers)
 {
 	assert(!d->server);
