@@ -42,7 +42,7 @@
 #include "zhttprequest.h"
 #include "statsmanager.h"
 #include "deferred.h"
-#include "httpserver.h"
+#include "simplehttpserver.h"
 #include "variantutil.h"
 #include "detectrule.h"
 #include "lastids.h"
@@ -64,6 +64,7 @@
 #define RETRY_WAIT_TIME 0
 #define WSCONTROL_WAIT_TIME 0
 #define STATE_RPC_TIMEOUT 1000
+#define DEFAULT_WS_KEEPALIVE_TIMEOUT 55
 
 using namespace VariantUtil;
 
@@ -214,7 +215,7 @@ public:
 			{
 				// determine session info
 				Deferred *d = SessionRequest::detectRulesGet(stateClient, requestData.uri.host().toUtf8(), requestData.uri.path(QUrl::FullyEncoded).toUtf8(), this);
-				connect(d, SIGNAL(finished(const DeferredResult &)), SLOT(sessionDetectRulesGet_finished(const DeferredResult &)));
+				connect(d, &Deferred::finished, this, &InspectWorker::sessionDetectRulesGet_finished);
 				return;
 			}
 
@@ -313,7 +314,7 @@ private slots:
 			if(!sid.isEmpty())
 			{
 				Deferred *d = SessionRequest::getLastIds(stateClient, sid, this);
-				connect(d, SIGNAL(finished(const DeferredResult &)), SLOT(sessionGetLastIds_finished(const DeferredResult &)));
+				connect(d, &Deferred::finished, this, &InspectWorker::sessionGetLastIds_finished);
 				return;
 			}
 		}
@@ -358,6 +359,7 @@ public:
 	HttpResponseData response;
 	ZhttpRequest *req;
 	QHash<QString, QString> meta;
+	bool debug;
 	bool autoCrossOrigin;
 	QByteArray jsonpCallback;
 	bool jsonpExtendedResponse;
@@ -366,24 +368,27 @@ public:
 	int timeout;
 	int keepAliveTimeout;
 	QByteArray keepAliveData;
+	bool responseSent;
 	QTimer *timer;
 	StatsManager *stats;
 
 	Hold(ZhttpRequest *_req, StatsManager *_stats, QObject *parent = 0) :
 		QObject(parent),
 		req(_req),
+		debug(false),
 		autoCrossOrigin(false),
 		jsonpExtendedResponse(false),
 		timeout(-1),
 		keepAliveTimeout(-1),
+		responseSent(false),
 		stats(_stats)
 	{
 		req->setParent(this);
-		connect(req, SIGNAL(bytesWritten(int)), SLOT(req_bytesWritten(int)));
-		connect(req, SIGNAL(error()), SLOT(req_error()));
+		connect(req, &ZhttpRequest::bytesWritten, this, &Hold::req_bytesWritten);
+		connect(req, &ZhttpRequest::error, this, &Hold::req_error);
 
 		timer = new QTimer(this);
-		connect(timer, SIGNAL(timeout()), SLOT(timer_timeout()));
+		connect(timer, &QTimer::timeout, this, &Hold::timer_timeout);
 	}
 
 	~Hold()
@@ -406,12 +411,15 @@ public:
 		}
 		else // StreamHold
 		{
-			// send initial response
-			response.headers.removeAll("Content-Length");
-			if(autoCrossOrigin)
-				Cors::applyCorsHeaders(requestData.headers, &response.headers);
-			req->beginResponse(response.code, response.reason, response.headers);
-			req->writeBody(response.body);
+			if(!responseSent)
+			{
+				// send initial response
+				response.headers.removeAll("Content-Length");
+				if(autoCrossOrigin)
+					Cors::applyCorsHeaders(requestData.headers, &response.headers);
+				req->beginResponse(response.code, response.reason, response.headers);
+				req->writeBody(response.body);
+			}
 
 			// start keep alive timer
 			if(keepAliveTimeout >= 0)
@@ -666,18 +674,29 @@ class WsSession : public QObject
 public:
 	QString cid;
 	QString channelPrefix;
+	HttpRequestData requestData;
+	QString route;
 	QString sid;
 	QHash<QString, QString> meta;
 	QHash<QString, QStringList> channelFilters; // k=channel, v=list(filters)
 	QSet<QString> channels;
 	int ttl;
+	QByteArray keepAliveType;
+	QByteArray keepAliveMessage;
 	QTimer *timer;
 
 	WsSession(QObject *parent = 0) :
 		QObject(parent)
 	{
 		timer = new QTimer(this);
-		connect(timer, SIGNAL(timeout()), SLOT(timer_timeout()));
+		connect(timer, &QTimer::timeout, this, &WsSession::timer_timeout);
+	}
+
+	~WsSession()
+	{
+		timer->disconnect(this);
+		timer->setParent(0);
+		timer->deleteLater();
 	}
 
 	void refreshExpiration()
@@ -816,6 +835,7 @@ public:
 	bool haveInspectInfo;
 	InspectData inspectInfo;
 	HttpResponseData responseData;
+	bool responseSent;
 	QString sid;
 	LastIds lastIds;
 	QList<Hold*> holds;
@@ -827,7 +847,8 @@ public:
 		cs(_cs),
 		zhttpIn(_zhttpIn),
 		stats(_stats),
-		haveInspectInfo(false)
+		haveInspectInfo(false),
+		responseSent(false)
 	{
 		req->setParent(this);
 	}
@@ -1073,6 +1094,17 @@ public:
 				haveInspectInfo = true;
 			}
 
+			if(args.contains("response-sent"))
+			{
+				if(args["response-sent"].type() != QVariant::Bool)
+				{
+					respondError("bad-request");
+					return;
+				}
+
+				responseSent = args["response-sent"].toBool();
+			}
+
 			bool useSession = false;
 			if(args.contains("use-session"))
 			{
@@ -1114,7 +1146,7 @@ public:
 				if(!rules.isEmpty())
 				{
 					Deferred *d = SessionRequest::detectRulesSet(stateClient, rules, this);
-					connect(d, SIGNAL(finished(const DeferredResult &)), SLOT(sessionDetectRulesSet_finished(const DeferredResult &)));
+					connect(d, &Deferred::finished, this, &AcceptWorker::sessionDetectRulesSet_finished);
 				}
 				else
 				{
@@ -1148,9 +1180,9 @@ signals:
 	void retryPacketReady(const RetryRequestPacket &packet);
 
 private:
-	void respondError(const QByteArray &condition)
+	void respondError(const QByteArray &condition, const QVariant &result = QVariant())
 	{
-		req->respondError(condition);
+		req->respondError(condition, result);
 		setFinished(true);
 	}
 
@@ -1159,7 +1191,7 @@ private:
 		if(!sid.isEmpty())
 		{
 			Deferred *d = SessionRequest::createOrUpdate(stateClient, sid, lastIds, this);
-			connect(d, SIGNAL(finished(const DeferredResult &)), SLOT(sessionCreateOrUpdate_finished(const DeferredResult &)));
+			connect(d, &Deferred::finished, this, &AcceptWorker::sessionCreateOrUpdate_finished);
 		}
 		else
 		{
@@ -1174,24 +1206,17 @@ private:
 		Instruct instruct = Instruct::fromResponse(responseData, &ok, &errorMessage);
 		if(!ok)
 		{
-			log_debug("failed to parse accept instructions: %s", qPrintable(errorMessage));
-
-			QVariantHash vresponse;
-			vresponse["code"] = 502;
-			vresponse["reason"] = QByteArray("Bad Gateway");
-			QVariantList vheaders;
-			vheaders += QVariant(QVariantList() << QByteArray("Content-Type") << QByteArray("text/plain"));
-			vresponse["headers"] = vheaders;
-			vresponse["body"] = QByteArray("Error while proxying to origin.\n");
-
-			QVariantHash result;
-			result["response"] = vresponse;
-			req->respond(result);
-
-			setFinished(true);
+			respondError("bad-format", errorMessage.toUtf8());
 			return;
 		}
 
+		// don't relay these headers. their meaning is handled by
+		//   zurl and they only apply to the outgoing hop.
+		instruct.response.headers.removeAll("Connection");
+		instruct.response.headers.removeAll("Keep-Alive");
+		instruct.response.headers.removeAll("Content-Encoding");
+		instruct.response.headers.removeAll("Transfer-Encoding");
+
 		if(instruct.holdMode == Instruct::NoHold)
 		{
 			QVariantHash vresponse;
@@ -1255,6 +1280,7 @@ private:
 					rpreq.rid = rs.rid;
 					rpreq.https = rs.isHttps;
 					rpreq.peerAddress = rs.peerAddress;
+					rpreq.debug = rs.debug;
 					rpreq.autoCrossOrigin = rs.autoCrossOrigin;
 					rpreq.jsonpCallback = rs.jsonpCallback;
 					rpreq.jsonpExtendedResponse = rs.jsonpExtendedResponse;
@@ -1303,6 +1329,7 @@ private:
 			ss.requestUri.setScheme(rs.isHttps ? "https" : "http");
 			ss.requestHeaders = requestData.headers;
 			ss.requestBody = requestData.body;
+			ss.responseCode = rs.responseCode;
 			ss.inSeq = rs.inSeq;
 			ss.outSeq = rs.outSeq;
 			ss.outCredits = rs.outCredits;
@@ -1317,6 +1344,7 @@ private:
 			hold->mode = instruct.holdMode;
 			hold->requestData = requestData;
 			hold->response = instruct.response;
+			hold->debug = rs.debug;
 			hold->autoCrossOrigin = rs.autoCrossOrigin;
 			hold->jsonpCallback = rs.jsonpCallback;
 			hold->jsonpExtendedResponse = rs.jsonpExtendedResponse;
@@ -1325,6 +1353,7 @@ private:
 			hold->keepAliveData = instruct.keepAliveData;
 			hold->sid = sid;
 			hold->meta = instruct.meta;
+			hold->responseSent = responseSent;
 
 			foreach(const Instruct::Channel &c, instruct.channels)
 				hold->channels.insert(channelPrefix + c.name, c);
@@ -1381,7 +1410,7 @@ public:
 	QZmq::Socket *statsSock;
 	QZmq::Socket *proxyStatsSock;
 	QZmq::Valve *proxyStatsValve;
-	HttpServer *controlHttpServer;
+	SimpleHttpServer *controlHttpServer;
 	StatsManager *stats;
 	CommonState cs;
 	QSet<Deferred*> deferreds;
@@ -1437,7 +1466,7 @@ public:
 			inspectServer = new ZrpcManager(this);
 			inspectServer->setBind(false);
 			inspectServer->setIpcFileMode(config.ipcFileMode);
-			connect(inspectServer, SIGNAL(requestReady()), SLOT(inspectServer_requestReady()));
+			connect(inspectServer, &ZrpcManager::requestReady, this, &Private::inspectServer_requestReady);
 
 			if(!inspectServer->setServerSpecs(QStringList() << config.inspectSpec))
 			{
@@ -1453,7 +1482,7 @@ public:
 			acceptServer = new ZrpcManager(this);
 			acceptServer->setBind(false);
 			acceptServer->setIpcFileMode(config.ipcFileMode);
-			connect(acceptServer, SIGNAL(requestReady()), SLOT(acceptServer_requestReady()));
+			connect(acceptServer, &ZrpcManager::requestReady, this, &Private::acceptServer_requestReady);
 
 			if(!acceptServer->setServerSpecs(QStringList() << config.acceptSpec))
 			{
@@ -1485,7 +1514,7 @@ public:
 			controlServer = new ZrpcManager(this);
 			controlServer->setBind(true);
 			controlServer->setIpcFileMode(config.ipcFileMode);
-			connect(controlServer, SIGNAL(requestReady()), SLOT(controlServer_requestReady()));
+			connect(controlServer, &ZrpcManager::requestReady, this, &Private::controlServer_requestReady);
 
 			if(!controlServer->setServerSpecs(QStringList() << config.commandSpec))
 			{
@@ -1509,7 +1538,7 @@ public:
 			}
 
 			inPullValve = new QZmq::Valve(inPullSock, this);
-			connect(inPullValve, SIGNAL(readyRead(const QList<QByteArray> &)), SLOT(inPull_readyRead(const QList<QByteArray> &)));
+			connect(inPullValve, &QZmq::Valve::readyRead, this, &Private::inPull_readyRead);
 
 			log_info("in pull: %s", qPrintable(config.pushInSpec));
 		}
@@ -1528,7 +1557,7 @@ public:
 			}
 
 			inSubValve = new QZmq::Valve(inSubSock, this);
-			connect(inSubValve, SIGNAL(readyRead(const QList<QByteArray> &)), SLOT(inSub_readyRead(const QList<QByteArray> &)));
+			connect(inSubValve, &QZmq::Valve::readyRead, this, &Private::inSub_readyRead);
 
 			log_info("in sub: %s", qPrintable(config.pushInSubSpec));
 		}
@@ -1562,7 +1591,7 @@ public:
 			}
 
 			wsControlInValve = new QZmq::Valve(wsControlInSock, this);
-			connect(wsControlInValve, SIGNAL(readyRead(const QList<QByteArray> &)), SLOT(wsControlIn_readyRead(const QList<QByteArray> &)));
+			connect(wsControlInValve, &QZmq::Valve::readyRead, this, &Private::wsControlIn_readyRead);
 
 			log_info("ws control in: %s", qPrintable(config.wsControlInSpec));
 
@@ -1580,8 +1609,8 @@ public:
 		}
 
 		stats = new StatsManager(this);
-		connect(stats, SIGNAL(connectionsRefreshed(const QList<QByteArray> &)), SLOT(stats_connectionsRefreshed(const QList<QByteArray> &)));
-		connect(stats, SIGNAL(unsubscribed(const QString &, const QString &)), SLOT(stats_unsubscribed(const QString &, const QString &)));
+		connect(stats, &StatsManager::connectionsRefreshed, this, &Private::stats_connectionsRefreshed);
+		connect(stats, &StatsManager::unsubscribed, this, &Private::stats_unsubscribed);
 
 		if(!config.statsSpec.isEmpty())
 		{
@@ -1612,7 +1641,7 @@ public:
 			}
 
 			proxyStatsValve = new QZmq::Valve(proxyStatsSock, this);
-			connect(proxyStatsValve, SIGNAL(readyRead(const QList<QByteArray> &)), SLOT(proxyStats_readyRead(const QList<QByteArray> &)));
+			connect(proxyStatsValve, &QZmq::Valve::readyRead, this, &Private::proxyStats_readyRead);
 
 			log_info("proxy stats: %s", qPrintable(config.proxyStatsSpec));
 		}
@@ -1633,8 +1662,8 @@ public:
 
 		if(config.pushInHttpPort != -1)
 		{
-			controlHttpServer = new HttpServer(this);
-			connect(controlHttpServer, SIGNAL(requestReady()), SLOT(controlHttpServer_requestReady()));
+			controlHttpServer = new SimpleHttpServer(this);
+			connect(controlHttpServer, &SimpleHttpServer::requestReady, this, &Private::controlHttpServer_requestReady);
 			controlHttpServer->listen(config.pushInHttpAddr, config.pushInHttpPort);
 
 			log_info("http control server: %s:%d", qPrintable(config.pushInHttpAddr.toString()), config.pushInHttpPort);
@@ -1783,15 +1812,37 @@ private:
 			{
 				WsControlPacket::Item i;
 				i.cid = s->cid.toUtf8();
-				i.type = WsControlPacket::Item::Send;
-				i.contentType = f.binary ? "binary" : "text";
-				i.message = f.body;
+
+				if(f.close)
+				{
+					i.type = WsControlPacket::Item::Close;
+					i.code = f.code;
+				}
+				else
+				{
+					i.type = WsControlPacket::Item::Send;
+
+					switch(f.messageType)
+					{
+						case PublishFormat::Text:   i.contentType = "text"; break;
+						case PublishFormat::Binary: i.contentType = "binary"; break;
+						case PublishFormat::Ping:   i.contentType = "ping"; break;
+						case PublishFormat::Pong:   i.contentType = "pong"; break;
+						default: continue; // unrecognized type, skip
+					}
+
+					i.message = f.body;
+				}
+
 				writeWsControlItem(i);
 			}
 
 			stats->addMessage(item.channel, item.id, "ws-message", wsSessions.count());
 		}
 
+		int receivers = responseHolds.count() + streamHolds.count() + wsSessions.count();
+		log_info("publish channel=%s receivers=%d", qPrintable(item.channel), receivers);
+
 		foreach(const QString &channel, responseUnsubs)
 			stats->removeSubscription("response", channel, true);
 
@@ -1810,7 +1861,7 @@ private:
 			}
 
 			Deferred *d = SessionRequest::updateMany(stateClient, sidLastIds, this);
-			connect(d, SIGNAL(finished(const DeferredResult &)), SLOT(sessionUpdateMany_finished(const DeferredResult &)));
+			connect(d, &Deferred::finished, this, &Private::sessionUpdateMany_finished);
 			deferreds += d;
 		}
 	}
@@ -1874,6 +1925,24 @@ private:
 		}
 	}
 
+	void httpControlRespond(SimpleHttpRequest *req, int code, const QByteArray &reason, const QString &body, const QByteArray &contentType = QByteArray(), const HttpHeaders &headers = HttpHeaders(), int items = -1)
+	{
+		HttpHeaders outHeaders = headers;
+		if(!contentType.isEmpty())
+			outHeaders += HttpHeader("Content-Type", contentType);
+		else
+			outHeaders += HttpHeader("Content-Type", "text/plain");
+
+		req->respond(code, reason, outHeaders, body.toUtf8());
+		connect(req, &SimpleHttpRequest::finished, req, &QObject::deleteLater);
+
+		QString msg = QString("control: %1 %2 code=%3 %4").arg(req->requestMethod(), QString::fromUtf8(req->requestUri()), QString::number(code), QString::number(body.size()));
+		if(items > -1)
+			msg += QString(" items=%1").arg(items);
+
+		log_info("%s", qPrintable(msg));
+	}
+
 private slots:
 	void inspectServer_requestReady()
 	{
@@ -1882,7 +1951,7 @@ private slots:
 			return;
 
 		InspectWorker *w = new InspectWorker(req, stateClient, config.shareAll, this);
-		connect(w, SIGNAL(finished(const DeferredResult &)), SLOT(inspectWorker_finished(const DeferredResult &)));
+		connect(w, &Deferred::finished, this, &Private::inspectWorker_finished);
 		deferreds += w;
 	}
 
@@ -1893,9 +1962,9 @@ private slots:
 			return;
 
 		AcceptWorker *w = new AcceptWorker(req, stateClient, &cs, zhttpIn, stats, this);
-		connect(w, SIGNAL(finished(const DeferredResult &)), SLOT(acceptWorker_finished(const DeferredResult &)));
-		connect(w, SIGNAL(holdsReady()), SLOT(acceptWorker_holdsReady()));
-		connect(w, SIGNAL(retryPacketReady(const RetryRequestPacket &)), SLOT(acceptWorker_retryPacketReady(const RetryRequestPacket &)));
+		connect(w, &AcceptWorker::finished, this, &Private::acceptWorker_finished);
+		connect(w, &AcceptWorker::holdsReady, this, &Private::acceptWorker_holdsReady);
+		connect(w, &AcceptWorker::retryPacketReady, this, &Private::acceptWorker_retryPacketReady);
 		deferreds += w;
 		w->start();
 	}
@@ -2026,14 +2095,16 @@ private slots:
 				if(!s)
 				{
 					s = new WsSession(this);
-					connect(s, SIGNAL(expired()), SLOT(wssession_expired()));
+					connect(s, &WsSession::expired, this, &Private::wssession_expired);
 					s->cid = QString::fromUtf8(item.cid);
 					s->ttl = item.ttl;
+					s->requestData.uri = item.uri;
 					s->refreshExpiration();
 					cs.wsSessions.insert(s->cid, s);
 					log_debug("added ws session: %s", qPrintable(s->cid));
 				}
 
+				s->route = item.route;
 				s->channelPrefix = QString::fromUtf8(item.channelPrefix);
 				continue;
 			}
@@ -2106,6 +2177,8 @@ private slots:
 
 					stats->addSubscription("ws", channel);
 					addSub(channel);
+
+					log_info("subscribe %s channel=%s", qPrintable(s->requestData.uri.toString(QUrl::FullyEncoded)), qPrintable(channel));
 				}
 				else if(cm.type == WsControlMessage::Unsubscribe)
 				{
@@ -2130,7 +2203,6 @@ private slots:
 					i.cid = item.cid;
 					i.type = WsControlPacket::Item::Detach;
 					writeWsControlItem(i);
-					continue;
 				}
 				else if(cm.type == WsControlMessage::Session)
 				{
@@ -2144,6 +2216,52 @@ private slots:
 					else
 						s->meta.remove(cm.metaName);
 				}
+				else if(cm.type == WsControlMessage::KeepAlive)
+				{
+					WsControlPacket::Item i;
+					i.cid = item.cid;
+					i.type = WsControlPacket::Item::KeepAliveSetup;
+
+					if(!cm.content.isNull())
+					{
+						QString contentType;
+						switch(cm.messageType)
+						{
+							case WsControlMessage::Text:   contentType = "text"; break;
+							case WsControlMessage::Binary: contentType = "binary"; break;
+							case WsControlMessage::Ping:   contentType = "ping"; break;
+							case WsControlMessage::Pong:   contentType = "pong"; break;
+							default: continue; // unrecognized type, ignore
+						}
+
+						s->keepAliveType = contentType.toUtf8();
+						s->keepAliveMessage = cm.content;
+
+						i.timeout = (cm.timeout > 0 ? cm.timeout : DEFAULT_WS_KEEPALIVE_TIMEOUT);
+					}
+					else
+					{
+						s->keepAliveType.clear();
+						s->keepAliveMessage.clear();
+					}
+
+					writeWsControlItem(i);
+				}
+			}
+			else if(item.type == WsControlPacket::Item::NeedKeepAlive)
+			{
+				if(!s->keepAliveMessage.isNull())
+				{
+					WsControlPacket::Item i;
+					i.cid = s->cid.toUtf8();
+					i.type = WsControlPacket::Item::Send;
+					i.contentType = s->keepAliveType;
+					i.message = s->keepAliveMessage;
+
+					writeWsControlItem(i);
+
+					stats->addActivity(s->route.toUtf8(), 1);
+				}
 			}
 		}
 
@@ -2152,7 +2270,7 @@ private slots:
 			foreach(const QString &sid, updateSids)
 			{
 				Deferred *d = SessionRequest::createOrUpdate(stateClient, sid, LastIds(), this);
-				connect(d, SIGNAL(finished(const DeferredResult &)), SLOT(sessionCreateOrUpdate_finished(const DeferredResult &)));
+				connect(d, &Deferred::finished, this, &Private::sessionCreateOrUpdate_finished);
 				deferreds += d;
 			}
 		}
@@ -2225,7 +2343,7 @@ private slots:
 				QHash<QString, LastIds> sidLastIds;
 				sidLastIds[sid] = LastIds();
 				Deferred *d = SessionRequest::updateMany(stateClient, sidLastIds, this);
-				connect(d, SIGNAL(finished(const DeferredResult &)), SLOT(sessionUpdateMany_finished(const DeferredResult &)));
+				connect(d, &Deferred::finished, this, &Private::sessionUpdateMany_finished);
 				deferreds += d;
 				return;
 			}
@@ -2234,11 +2352,64 @@ private slots:
 
 	void controlHttpServer_requestReady()
 	{
-		HttpRequest *req = controlHttpServer->takeNext();
+		SimpleHttpRequest *req = controlHttpServer->takeNext();
 		if(!req)
 			return;
 
-		if(req->requestUri() == "/publish/" || req->requestUri() == "/publish")
+		QByteArray path = req->requestUri();
+		if(path.endsWith("/"))
+			path.truncate(path.length() - 1);
+
+		HttpHeaders headers = req->requestHeaders();
+
+		QByteArray responseContentType;
+		if(headers.contains("Accept"))
+		{
+			foreach(const HttpHeaderParameters &params, headers.getAllAsParameters("Accept"))
+			{
+				if(params.isEmpty() || params[0].first.isEmpty())
+					continue;
+
+				QList<QByteArray> type = params[0].first.split('/');
+				if(type[0] == "text" && type[1] == "plain")
+				{
+					responseContentType = "text/plain";
+					break;
+				}
+				else if(type[0] == "application" && type[1] == "json")
+				{
+					responseContentType = "application/json";
+					break;
+				}
+				else if(type[0] == "text" && type[1] == "*")
+				{
+					responseContentType = "text/plain";
+					break;
+				}
+				else if(type[0] == "application" && type[1] == "*")
+				{
+					responseContentType = "application/json";
+					break;
+				}
+				else if(type[0] == "*" && type[1] == "*")
+				{
+					responseContentType = "text/plain";
+					break;
+				}
+			}
+
+			if(responseContentType.isEmpty())
+			{
+				httpControlRespond(req, 406, "Not Acceptable", "Not Acceptable. Supported formats are text/plain and application/json.\n");
+				return;
+			}
+		}
+		else
+		{
+			responseContentType = "text/plain";
+		}
+
+		if(path == "/publish")
 		{
 			if(req->requestMethod() == "POST")
 			{
@@ -2246,15 +2417,13 @@ private slots:
 				QJsonDocument doc = QJsonDocument::fromJson(req->requestBody(), &e);
 				if(e.error != QJsonParseError::NoError)
 				{
-					req->respond(400, "Bad Request", "Body is not valid JSON.\n");
-					connect(req, SIGNAL(finished()), req, SLOT(deleteLater()));
+					httpControlRespond(req, 400, "Bad Request", "Body is not valid JSON.\n");
 					return;
 				}
 
 				if(!doc.isObject())
 				{
-					req->respond(400, "Bad Request", "Invalid format.\n");
-					connect(req, SIGNAL(finished()), req, SLOT(deleteLater()));
+					httpControlRespond(req, 400, "Bad Request", "Invalid format.\n");
 					return;
 				}
 
@@ -2263,15 +2432,13 @@ private slots:
 
 				if(!mdata.contains("items"))
 				{
-					req->respond(400, "Bad Request", "Invalid format: object does not contain 'items'\n");
-					connect(req, SIGNAL(finished()), req, SLOT(deleteLater()));
+					httpControlRespond(req, 400, "Bad Request", "Invalid format: object does not contain 'items'\n");
 					return;
 				}
 
 				if(mdata["items"].type() != QVariant::List)
 				{
-					req->respond(400, "Bad Request", "Invalid format: object contains 'items' with wrong type\n");
-					connect(req, SIGNAL(finished()), req, SLOT(deleteLater()));
+					httpControlRespond(req, 400, "Bad Request", "Invalid format: object contains 'items' with wrong type\n");
 					return;
 				}
 
@@ -2282,30 +2449,37 @@ private slots:
 				QList<PublishItem> items = parseHttpItems(vitems, &ok, &errorMessage);
 				if(!ok)
 				{
-					req->respond(400, "Bad Request", QString("Invalid format: %1\n").arg(errorMessage));
-					connect(req, SIGNAL(finished()), req, SLOT(deleteLater()));
+					httpControlRespond(req, 400, "Bad Request", QString("Invalid format: %1\n").arg(errorMessage));
 					return;
 				}
 
+				QString message = "Published";
+				if(responseContentType == "application/json")
+				{
+					QVariantMap obj;
+					obj["message"] = message;
+					QString body = QJsonDocument(QJsonObject::fromVariantMap(obj)).toJson(QJsonDocument::Compact);
+					httpControlRespond(req, 200, "OK", body + "\n", responseContentType, HttpHeaders(), items.count());
+				}
+				else // text/plain
+				{
+					httpControlRespond(req, 200, "OK", message + "\n", responseContentType, HttpHeaders(), items.count());
+				}
+
 				foreach(const PublishItem &item, items)
 					handlePublishItem(item);
-
-				req->respond(200, "OK", "Published\n");
 			}
 			else
 			{
 				HttpHeaders headers;
-				headers += HttpHeader("Content-Type", "text/plain");
 				headers += HttpHeader("Allow", "POST");
-				req->respond(405, "Method Not Allowed", headers, "Method not allowed: " + req->requestMethod().toUtf8() + ".\n");
+				httpControlRespond(req, 405, "Method Not Allowed", "Method not allowed: " + req->requestMethod() + ".\n", QByteArray(), headers);
 			}
 		}
 		else
 		{
-			req->respond(404, "Not Found", "Not Found\n");
+			httpControlRespond(req, 404, "Not Found", "Not Found\n");
 		}
-
-		connect(req, SIGNAL(finished()), req, SLOT(deleteLater()));
 	}
 
 	void sessionCreateOrUpdate_finished(const DeferredResult &result)
@@ -2352,7 +2526,7 @@ private slots:
 		foreach(Hold *hold, holds)
 		{
 			hold->setParent(this);
-			connect(hold, SIGNAL(finished()), SLOT(hold_finished()));
+			connect(hold, &Hold::finished, this, &Private::hold_finished);
 			cs.holds.insert(hold->req->rid(), hold);
 
 			QHashIterator<QString, Instruct::Channel> it(hold->channels);
@@ -2380,6 +2554,8 @@ private slots:
 					cs.streamHoldsByChannel[channel] += hold;
 				}
 
+				log_info("subscribe %s channel=%s", qPrintable(hold->requestData.uri.toString(QUrl::FullyEncoded)), qPrintable(channel));
+
 				stats->addSubscription(hold->mode == Instruct::ResponseHold ? "response" : "stream", channel);
 				addSub(channel);
 			}
@@ -2448,7 +2624,7 @@ private slots:
 			if(!sidLastIds.isEmpty())
 			{
 				Deferred *d = SessionRequest::updateMany(stateClient, sidLastIds, this);
-				connect(d, SIGNAL(finished(const DeferredResult &)), SLOT(sessionUpdateMany_finished(const DeferredResult &)));
+				connect(d, &Deferred::finished, this, &Private::sessionUpdateMany_finished);
 				deferreds += d;
 			}
 		}
