@@ -29,21 +29,25 @@ class RequestState
 {
 public:
 	ZhttpRequest::Rid rid;
+	int responseCode;
 	int inSeq;
 	int outSeq;
 	int outCredits;
 	QHostAddress peerAddress;
 	bool isHttps;
+	bool debug;
 	bool autoCrossOrigin;
 	QByteArray jsonpCallback;
 	bool jsonpExtendedResponse;
 	QVariant userData;
 
 	RequestState() :
+		responseCode(-1),
 		inSeq(0),
 		outSeq(0),
 		outCredits(0),
 		isHttps(false),
+		debug(false),
 		autoCrossOrigin(false),
 		jsonpExtendedResponse(false)
 	{
