@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2012-2015 Fanout, Inc.
+ * Copyright (C) 2012-2016 Fanout, Inc.
  *
  * This file is part of Pushpin.
  *
@@ -164,12 +164,20 @@ public:
 	{
 		config = _config;
 
-		domainMap = new DomainMap(config.routesFile);
-		connect(domainMap, SIGNAL(changed()), SLOT(domainMap_changed()));
+		if(!config.routeLines.isEmpty())
+		{
+			domainMap = new DomainMap(this);
+			foreach(const QString &line, config.routeLines)
+				domainMap->addRouteLine(line);
+		}
+		else
+			domainMap = new DomainMap(config.routesFile, this);
+
+		connect(domainMap, &DomainMap::changed, this, &Private::domainMap_changed);
 
 		zhttpIn = new ZhttpManager(this);
-		connect(zhttpIn, SIGNAL(requestReady()), SLOT(zhttpIn_requestReady()));
-		connect(zhttpIn, SIGNAL(socketReady()), SLOT(zhttpIn_socketReady()));
+		connect(zhttpIn, &ZhttpManager::requestReady, this, &Private::zhttpIn_requestReady);
+		connect(zhttpIn, &ZhttpManager::socketReady, this, &Private::zhttpIn_socketReady);
 
 		zhttpIn->setInstanceId(config.clientId);
 		zhttpIn->setServerInSpecs(config.serverInSpecs);
@@ -183,7 +191,7 @@ public:
 		zroutes->setDefaultInSpecs(config.clientInSpecs);
 
 		sockJsManager = new SockJsManager(config.sockJsUrl, this);
-		connect(sockJsManager, SIGNAL(sessionReady()), SLOT(sockjs_sessionReady()));
+		connect(sockJsManager, &SockJsManager::sessionReady, this, &Private::sockjs_sessionReady);
 
 		if(!config.inspectSpec.isEmpty())
 		{
@@ -230,7 +238,7 @@ public:
 			}
 
 			handler_retry_in_valve = new QZmq::Valve(handler_retry_in_sock, this);
-			connect(handler_retry_in_valve, SIGNAL(readyRead(const QList<QByteArray> &)), SLOT(handler_retry_in_readyRead(const QList<QByteArray> &)));
+			connect(handler_retry_in_valve, &QZmq::Valve::readyRead, this, &Private::handler_retry_in_readyRead);
 		}
 
 		if(handler_retry_in_valve)
@@ -274,7 +282,7 @@ public:
 			command = new ZrpcManager(this);
 			command->setBind(true);
 			command->setIpcFileMode(config.ipcFileMode);
-			connect(command, SIGNAL(requestReady()), SLOT(command_requestReady()));
+			connect(command, &ZrpcManager::requestReady, this, &Private::command_requestReady);
 
 			if(!command->setServerSpecs(QStringList() << config.commandSpec))
 			{
@@ -285,7 +293,7 @@ public:
 
 		if(!config.appVersion.isEmpty() && config.updatesCheck)
 		{
-			new Updater(config.appVersion, zroutes->defaultManager(), this);
+			new Updater(config.appVersion, config.organizationName, zroutes->defaultManager(), this);
 		}
 
 		// init zroutes
@@ -323,16 +331,18 @@ public:
 			log_debug("creating proxysession for id=%s", rs->rid().second.data());
 
 			ps = new ProxySession(zroutes, accept, this);
-			connect(ps, SIGNAL(addNotAllowed()), SLOT(ps_addNotAllowed()));
-			connect(ps, SIGNAL(finished()), SLOT(ps_finished()));
-			connect(ps, SIGNAL(requestSessionDestroyed(RequestSession *, bool)), SLOT(ps_requestSessionDestroyed(RequestSession *, bool)));
+			connect(ps, &ProxySession::addNotAllowed, this, &Private::ps_addNotAllowed);
+			connect(ps, &ProxySession::finished, this, &Private::ps_finished);
+			connect(ps, &ProxySession::requestSessionDestroyed, this, &Private::ps_requestSessionDestroyed);
 
 			ps->setRoute(route);
 			ps->setDefaultSigKey(config.sigIss, config.sigKey);
 			ps->setDefaultUpstreamKey(config.upstreamKey);
+			ps->setAcceptXForwardedProtocol(config.acceptXForwardedProtocol);
 			ps->setUseXForwardedProtocol(config.useXForwardedProtocol);
 			ps->setXffRules(config.xffUntrustedRule, config.xffTrustedRule);
 			ps->setOrigHeadersNeedMark(config.origHeadersNeedMark);
+			ps->setProxyInitialResponseEnabled(true);
 
 			if(idata)
 				ps->setInspectData(*idata);
@@ -368,10 +378,11 @@ public:
 		QByteArray cid = connectionManager.addConnection(sock);
 
 		WsProxySession *ps = new WsProxySession(zroutes, &connectionManager, stats, wsControl, this);
-		connect(ps, SIGNAL(finishedByPassthrough()), SLOT(wsps_finishedByPassthrough()));
+		connect(ps, &WsProxySession::finishedByPassthrough, this, &Private::wsps_finishedByPassthrough);
 
 		ps->setDefaultSigKey(config.sigIss, config.sigKey);
 		ps->setDefaultUpstreamKey(config.upstreamKey);
+		ps->setAcceptXForwardedProtocol(config.acceptXForwardedProtocol);
 		ps->setUseXForwardedProtocol(config.useXForwardedProtocol);
 		ps->setXffRules(config.xffUntrustedRule, config.xffTrustedRule);
 		ps->setOrigHeadersNeedMark(config.origHeadersNeedMark);
@@ -403,6 +414,12 @@ public:
 		return true;
 	}
 
+	bool isXForwardedProtocolTls(const HttpHeaders &headers)
+	{
+		QByteArray xfp = headers.get("X-Forwarded-Protocol");
+		return (!xfp.isEmpty() && (xfp == "https" || xfp == "wss"));
+	}
+
 	void tryTakeRequest()
 	{
 		if(!canTake())
@@ -412,12 +429,16 @@ public:
 		if(!req)
 			return;
 
-		RequestSession *rs = new RequestSession(domainMap, sockJsManager, inspect, inspectChecker, accept, this);
-		connect(rs, SIGNAL(inspected(const InspectData &)), SLOT(rs_inspected(const InspectData &)));
-		connect(rs, SIGNAL(inspectError()), SLOT(rs_inspectError()));
-		connect(rs, SIGNAL(finished()), SLOT(rs_finished()));
-		connect(rs, SIGNAL(finishedByAccept()), SLOT(rs_finishedByAccept()));
+		if(config.acceptXForwardedProtocol && isXForwardedProtocolTls(req->requestHeaders()))
+			req->setIsTls(true);
 
+		RequestSession *rs = new RequestSession(domainMap, sockJsManager, inspect, inspectChecker, accept, this);
+		connect(rs, &RequestSession::inspected, this, &Private::rs_inspected);
+		connect(rs, &RequestSession::inspectError, this, &Private::rs_inspectError);
+		connect(rs, &RequestSession::finished, this, &Private::rs_finished);
+		connect(rs, &RequestSession::finishedByAccept, this, &Private::rs_finishedByAccept);
+
+		rs->setDebugEnabled(config.debug);
 		rs->setAutoCrossOrigin(config.autoCrossOrigin);
 		rs->setPrefetchSize(config.inspectPrefetch);
 
@@ -435,9 +456,12 @@ public:
 		if(!sock)
 			return;
 
+		if(config.acceptXForwardedProtocol && isXForwardedProtocolTls(sock->requestHeaders()))
+			sock->setIsTls(true);
+
 		QUrl requestUri = sock->requestUri();
 
-		log_info("IN ws id=%s, %s", sock->rid().second.data(), requestUri.toEncoded().data());
+		log_debug("IN ws id=%s, %s", sock->rid().second.data(), requestUri.toEncoded().data());
 
 		bool isSecure = (requestUri.scheme() == "wss");
 		QString host = requestUri.host();
@@ -467,7 +491,7 @@ public:
 		if(!sock)
 			return;
 
-		log_info("IN sockjs obj=%p %s", sock, sock->requestUri().toEncoded().data());
+		log_debug("IN sockjs obj=%p %s", sock, sock->requestUri().toEncoded().data());
 
 		log_debug("creating wsproxysession for sockjs=%p", sock);
 		doProxySocket(sock, sock->route());
@@ -480,6 +504,36 @@ public:
 		tryTakeSockJsSession();
 	}
 
+	void logFinished(RequestSession *rs, bool accepted = false)
+	{
+		HttpRequestData rd = rs->requestData();
+		DomainMap::Entry e = rs->route();
+
+		QString msg = QString("%1 %2").arg(rd.method, rd.uri.toString(QUrl::FullyEncoded));
+		QUrl ref = QUrl(QString::fromUtf8(rd.headers.get("Referer")));
+		if(!ref.isEmpty())
+			msg += QString(" ref=%1").arg(ref.toString(QUrl::FullyEncoded));
+		if(!e.id.isEmpty())
+			msg += QString(" route=%1").arg(QString::fromUtf8(e.id));
+
+		HttpResponseData resp = rs->responseData();
+
+		if(accepted)
+		{
+			msg += " hold";
+		}
+		else if(resp.code != -1)
+		{
+			msg += QString(" code=%1 %2").arg(QString::number(resp.code), QString::number(rs->responseBodySize()));
+		}
+		else
+		{
+			msg += " error";
+		}
+
+		log_info("%s", qPrintable(msg));
+	}
+
 private slots:
 	void zhttpIn_requestReady()
 	{
@@ -519,6 +573,8 @@ private slots:
 	{
 		RequestSession *rs = (RequestSession *)sender();
 
+		logFinished(rs);
+
 		if(stats)
 			stats->removeConnection(ridToString(rs->rid()), false);
 
@@ -532,6 +588,8 @@ private slots:
 	{
 		RequestSession *rs = (RequestSession *)sender();
 
+		logFinished(rs, true);
+
 		if(stats)
 		{
 			// add connection so that it becomes lingerable
@@ -632,7 +690,7 @@ private slots:
 			return;
 		}
 
-		log_info("IN (retry) %s %s", qPrintable(p.requestData.method), p.requestData.uri.toEncoded().data());
+		log_debug("IN (retry) %s %s", qPrintable(p.requestData.method), p.requestData.uri.toEncoded().data());
 
 		InspectData idata;
 		if(p.haveInspectInfo)
@@ -666,7 +724,7 @@ private slots:
 			// note: if the routing table was changed, there's a chance the request
 			//   might get a different route id this time around. this could confuse
 			//   stats processors tracking route+connection mappings.
-			rs->startRetry(zhttpRequest, req.autoCrossOrigin, req.jsonpCallback, req.jsonpExtendedResponse);
+			rs->startRetry(zhttpRequest, req.debug, req.autoCrossOrigin, req.jsonpCallback, req.jsonpExtendedResponse);
 
 			doProxy(rs, p.haveInspectInfo ? &idata : 0, true);
 		}
