@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2012-2015 Fanout, Inc.
+ * Copyright (C) 2012-2016 Fanout, Inc.
  *
  * This file is part of Pushpin.
  *
@@ -33,10 +33,12 @@
 #include "zhttpmanager.h"
 #include "zhttprequest.h"
 #include "zroutes.h"
+#include "statusreasons.h"
 #include "xffrule.h"
 #include "requestsession.h"
 #include "proxyutil.h"
 #include "acceptrequest.h"
+#include "testhttprequest.h"
 
 #define MAX_ACCEPT_REQUEST_BODY 100000
 #define MAX_ACCEPT_RESPONSE_BODY 100000
@@ -73,11 +75,15 @@ public:
 
 		RequestSession *rs;
 		State state;
+		bool startedResponse;
+		bool unclean;
 		int bytesToWrite;
 
 		SessionItem() :
 			rs(0),
 			state(WaitingForResponse),
+			startedResponse(false),
+			unclean(false),
 			bytesToWrite(0)
 		{
 		}
@@ -92,15 +98,18 @@ public:
 	bool isHttps;
 	DomainMap::Entry route;
 	QList<DomainMap::Target> targets;
-	ZhttpRequest *zhttpRequest;
+	DomainMap::Target target;
+	HttpRequest *zhttpRequest;
 	bool addAllowed;
 	bool haveInspectData;
 	InspectData idata;
 	QSet<QByteArray> acceptHeaderPrefixes;
 	QSet<QByteArray> acceptContentTypes;
 	QSet<SessionItem*> sessionItems;
+	bool shared;
 	HttpRequestData requestData;
 	HttpResponseData responseData;
+	HttpResponseData acceptResponseData;
 	BufferList requestBody;
 	BufferList responseBody;
 	QHash<RequestSession*, SessionItem*> sessionItemsBySession;
@@ -113,10 +122,13 @@ public:
 	QByteArray defaultSigKey;
 	QByteArray defaultUpstreamKey;
 	bool passToUpstream;
+	bool acceptXForwardedProtocol;
 	bool useXForwardedProtocol;
 	XffRule xffRule;
 	XffRule xffTrustedRule;
 	QList<QByteArray> origHeadersNeedMark;
+	bool proxyInitialResponse;
+	bool acceptAfterResponding;
 	AcceptRequest *acceptRequest;
 
 	Private(ProxySession *_q, ZRoutes *_zroutes, ZrpcManager *_acceptManager) :
@@ -131,11 +143,15 @@ public:
 		zhttpRequest(0),
 		addAllowed(true),
 		haveInspectData(false),
+		shared(false),
 		requestBytesToWrite(0),
 		requestBodySent(false),
 		total(0),
 		passToUpstream(false),
+		acceptXForwardedProtocol(false),
 		useXForwardedProtocol(false),
+		proxyInitialResponse(false),
+		acceptAfterResponding(false),
 		acceptRequest(0)
 	{
 		acceptHeaderPrefixes += "Grip-";
@@ -175,12 +191,16 @@ public:
 		SessionItem *si = new SessionItem;
 		si->rs = rs;
 		si->rs->setParent(this);
+
+		if(!sessionItems.isEmpty())
+			shared = true;
+
 		sessionItems += si;
 		sessionItemsBySession.insert(rs, si);
-		connect(rs, SIGNAL(bytesWritten(int)), SLOT(rs_bytesWritten(int)));
-		connect(rs, SIGNAL(errorResponding()), SLOT(rs_errorResponding()));
-		connect(rs, SIGNAL(finished()), SLOT(rs_finished()));
-		connect(rs, SIGNAL(paused()), SLOT(rs_paused()));
+		connect(rs, &RequestSession::bytesWritten, this, &Private::rs_bytesWritten);
+		connect(rs, &RequestSession::errorResponding, this, &Private::rs_errorResponding);
+		connect(rs, &RequestSession::finished, this, &Private::rs_finished);
+		connect(rs, &RequestSession::paused, this, &Private::rs_paused);
 
 		if(state == Stopped)
 		{
@@ -218,7 +238,7 @@ public:
 
 			targets = route.targets;
 
-			bool trustedClient = ProxyUtil::manipulateRequestHeaders("wsproxysession", q, &requestData, defaultUpstreamKey, route, sigIss, sigKey, useXForwardedProtocol, xffTrustedRule, xffRule, origHeadersNeedMark, rs->peerAddress(), idata);
+			bool trustedClient = ProxyUtil::manipulateRequestHeaders("wsproxysession", q, &requestData, defaultUpstreamKey, route, sigIss, sigKey, acceptXForwardedProtocol, useXForwardedProtocol, xffTrustedRule, xffRule, origHeadersNeedMark, rs->peerAddress(), idata);
 
 			if(trustedClient)
 				passToUpstream = true;
@@ -229,8 +249,8 @@ public:
 			if(!rs->isRetry())
 			{
 				inRequest = rs->request();
-				connect(inRequest, SIGNAL(readyRead()), SLOT(inRequest_readyRead()));
-				connect(inRequest, SIGNAL(error()), SLOT(inRequest_error()));
+				connect(inRequest, &ZhttpRequest::readyRead, this, &Private::inRequest_readyRead);
+				connect(inRequest, &ZhttpRequest::error, this, &Private::inRequest_error);
 
 				requestBody += inRequest->readBody();
 			}
@@ -254,6 +274,7 @@ public:
 			// get the session caught up with where we're at
 
 			si->state = SessionItem::Responding;
+			si->startedResponse = true;
 			rs->startResponse(responseData.code, responseData.reason, responseData.headers);
 
 			if(!responseBody.isEmpty())
@@ -279,11 +300,18 @@ public:
 	{
 		if(targets.isEmpty())
 		{
-			rejectAll(502, "Bad Gateway", "Error while proxying to origin.");
+			QString msg = "Error while proxying to origin.";
+
+			QStringList targetStrs;
+			foreach(const DomainMap::Target &t, route.targets)
+				targetStrs += targetToString(t);
+			QString dmsg = QString("Unable to connect to any targets. Tried: %1").arg(targetStrs.join(", "));
+
+			rejectAll(502, "Bad Gateway", msg, dmsg);
 			return;
 		}
 
-		DomainMap::Target target = targets.takeFirst();
+		target = targets.takeFirst();
 
 		if(target.overHttp)
 		{
@@ -296,7 +324,7 @@ public:
 
 			if(contentType == "application/websocket-events" && !passToUpstream)
 			{
-				rejectAll(502, "Bad Gateway", "Error while proxying to origin.");
+				rejectAll(403, "Forbidden", "Client not allowed to send WebSocket events directly.");
 				return;
 			}
 		}
@@ -311,26 +339,48 @@ public:
 			uri.setHost(target.host);
 
 		if(zhttpManager)
+		{
 			zroutes->removeRef(zhttpManager);
-
-		if(target.type == DomainMap::Target::Custom)
-		{
-			zhttpManager = zroutes->managerForRoute(target.zhttpRoute);
-			log_debug("proxysession: %p forwarding to %s", q, qPrintable(target.zhttpRoute.baseSpec));
-		}
-		else // Default
-		{
-			zhttpManager = zroutes->defaultManager();
-			log_debug("proxysession: %p forwarding to %s:%d", q, qPrintable(target.connectHost), target.connectPort);
+			zhttpManager = 0;
 		}
 
-		zroutes->addRef(zhttpManager);
+		if(target.type == DomainMap::Target::Test)
+		{
+			// for test route, auto-adjust path
+			if(!route.pathBeg.isEmpty())
+			{
+				int pathRemove = route.pathBeg.length();
+				if(route.pathBeg.endsWith('/'))
+					--pathRemove;
 
-		zhttpRequest = zhttpManager->createRequest();
-		zhttpRequest->setParent(this);
-		connect(zhttpRequest, SIGNAL(readyRead()), SLOT(zhttpRequest_readyRead()));
-		connect(zhttpRequest, SIGNAL(bytesWritten(int)), SLOT(zhttpRequest_bytesWritten(int)));
-		connect(zhttpRequest, SIGNAL(error()), SLOT(zhttpRequest_error()));
+				if(pathRemove > 0)
+					uri.setPath(uri.path(QUrl::FullyEncoded).mid(pathRemove));
+			}
+
+			zhttpRequest = new TestHttpRequest(this);
+		}
+		else
+		{
+			if(target.type == DomainMap::Target::Custom)
+			{
+				zhttpManager = zroutes->managerForRoute(target.zhttpRoute);
+				log_debug("proxysession: %p forwarding to %s", q, qPrintable(target.zhttpRoute.baseSpec));
+			}
+			else // Default
+			{
+				zhttpManager = zroutes->defaultManager();
+				log_debug("proxysession: %p forwarding to %s:%d", q, qPrintable(target.connectHost), target.connectPort);
+			}
+
+			zroutes->addRef(zhttpManager);
+
+			zhttpRequest = zhttpManager->createRequest();
+			zhttpRequest->setParent(this);
+		}
+
+		connect(zhttpRequest, &ZhttpRequest::readyRead, this, &Private::zhttpRequest_readyRead);
+		connect(zhttpRequest, &ZhttpRequest::bytesWritten, this, &Private::zhttpRequest_bytesWritten);
+		connect(zhttpRequest, &ZhttpRequest::error, this, &Private::zhttpRequest_error);
 
 		if(target.trusted)
 			zhttpRequest->setIgnorePolicies(true);
@@ -424,16 +474,55 @@ public:
 		{
 			if(si->state != SessionItem::Errored)
 			{
-				assert(si->state == SessionItem::WaitingForResponse);
+				if(si->state == SessionItem::Paused)
+				{
+					if(si->startedResponse)
+						si->state = SessionItem::Responding;
+					else
+						si->state = SessionItem::WaitingForResponse;
 
-				si->state = SessionItem::Responded;
-				si->bytesToWrite = -1;
-				si->rs->respondCannotAccept();
+					si->rs->resume();
+				}
+
+				assert(si->state == SessionItem::WaitingForResponse || si->state == SessionItem::Responding);
+
+				if(si->state == SessionItem::WaitingForResponse)
+				{
+					si->state = SessionItem::Responded;
+					si->bytesToWrite = -1;
+
+					si->rs->respondCannotAccept();
+				}
+				else
+				{
+					// if we already started responding, then only provide an
+					//   error message in debug mode
+
+					if(si->rs->debugEnabled())
+					{
+						// if debug enabled, append the message at the end.
+						//   this may ruin the content, but hey it's debug
+						//   mode
+						QByteArray buf = "\n\nAccept service unavailable\n";
+						si->bytesToWrite += buf.size();
+						si->rs->writeResponseBody(buf);
+						si->rs->endResponseBody();
+					}
+					else
+					{
+						// if debug not enabled, then the best we can do is
+						//   disconnect
+						si->state = SessionItem::Responded;
+						si->unclean = true;
+						si->bytesToWrite = -1;
+						si->rs->endResponseBody();
+					}
+				}
 			}
 		}
 	}
 
-	void rejectAll(int code, const QString &reason, const QString &errorMessage)
+	void rejectAll(int code, const QString &reason, const QString &errorMessage, const QString &debugErrorMessage)
 	{
 		// kill the active target request, if any
 		delete zhttpRequest;
@@ -448,15 +537,59 @@ public:
 		{
 			if(si->state != SessionItem::Errored)
 			{
-				assert(si->state == SessionItem::WaitingForResponse);
+				if(si->state == SessionItem::Paused)
+				{
+					if(si->startedResponse)
+						si->state = SessionItem::Responding;
+					else
+						si->state = SessionItem::WaitingForResponse;
 
-				si->state = SessionItem::Responded;
-				si->bytesToWrite = -1;
-				si->rs->respondError(code, reason, errorMessage);
+					si->rs->resume();
+				}
+
+				assert(si->state == SessionItem::WaitingForResponse || si->state == SessionItem::Responding);
+
+				if(si->state == SessionItem::WaitingForResponse)
+				{
+					si->state = SessionItem::Responded;
+					si->bytesToWrite = -1;
+
+					si->rs->respondError(code, reason, si->rs->debugEnabled() ? debugErrorMessage : errorMessage);
+				}
+				else // Responding
+				{
+					// if we already started responding, then only provide a
+					//   rejection message in debug mode
+
+					if(si->rs->debugEnabled())
+					{
+						// if debug enabled, append the message at the end.
+						//   this may ruin the content, but hey it's debug
+						//   mode
+						QByteArray buf = "\n\n" + debugErrorMessage.toUtf8() + '\n';
+						si->bytesToWrite += buf.size();
+						si->rs->writeResponseBody(buf);
+						si->rs->endResponseBody();
+					}
+					else
+					{
+						// if debug not enabled, then the best we can do is
+						//   disconnect
+						si->state = SessionItem::Responded;
+						si->unclean = true;
+						si->bytesToWrite = -1;
+						si->rs->endResponseBody();
+					}
+				}
 			}
 		}
 	}
 
+	void rejectAll(int code, const QString &reason, const QString &errorMessage)
+	{
+		rejectAll(code, reason, errorMessage, errorMessage);
+	}
+
 	void respondAll(int code, const QByteArray &reason, const HttpHeaders &headers, const QByteArray &body)
 	{
 		assert(state != Responding);
@@ -479,18 +612,22 @@ public:
 
 	void destroyAll()
 	{
-		// this method is only to be called when we are in Responding state
-		assert(state == Responding);
+		assert(state == Accepting || state == Responding);
 
 		state = Responded;
 
 		foreach(SessionItem *si, sessionItems)
 		{
-			assert(si->state != SessionItem::WaitingForResponse);
+			if(si->state == SessionItem::Paused)
+			{
+				si->state = SessionItem::Responding;
+				si->rs->resume();
+			}
 
-			if(si->state == SessionItem::Responding)
+			if(si->state == SessionItem::WaitingForResponse || si->state == SessionItem::Responding)
 			{
 				si->state = SessionItem::Responded;
+				si->unclean = true;
 				si->bytesToWrite = -1;
 				si->rs->endResponseBody();
 			}
@@ -517,7 +654,10 @@ public:
 			{
 				if(responseBody.size() + buf.size() > MAX_ACCEPT_RESPONSE_BODY)
 				{
-					rejectAll(502, "Bad Gateway", "GRIP instruct response too large.");
+					QString msg = "Error while proxying to origin.";
+					QString dmsg = QString("GRIP instruct response too large from %1").arg(targetToString(target));
+
+					rejectAll(502, "Bad Gateway", msg, dmsg);
 					return;
 				}
 
@@ -591,8 +731,10 @@ public:
 					return;
 			}
 
-			if(state == Accepting)
+			if(state == Accepting || (state == Responding && acceptAfterResponding))
 			{
+				state = Accepting;
+
 				if(acceptManager)
 				{
 					log_debug("we have an acceptmanager");
@@ -603,7 +745,9 @@ public:
 					}
 				}
 				else
+				{
 					cannotAcceptAll();
+				}
 			}
 			else if(state == Responding)
 			{
@@ -621,6 +765,55 @@ public:
 		}
 	}
 
+	static QString targetToString(const DomainMap::Target &target)
+	{
+		if(target.type == DomainMap::Target::Test)
+			return "test";
+		else if(target.type == DomainMap::Target::Custom)
+			return(target.zhttpRoute.req ? "zhttpreq/" : "zhttp/") + target.zhttpRoute.baseSpec;
+		else // Default
+			return target.connectHost + ':' + QString::number(target.connectPort);
+	}
+
+	void logFinished(SessionItem *si, bool accepted = false)
+	{
+		RequestSession *rs = si->rs;
+
+		HttpRequestData rd = rs->requestData();
+
+		QString msg = QString("%1 %2 -> %3").arg(rd.method, rd.uri.toString(QUrl::FullyEncoded), targetToString(target));
+
+		QUrl ref = QUrl(QString::fromUtf8(rd.headers.get("Referer")));
+		if(!ref.isEmpty())
+			msg += QString(" ref=%1").arg(ref.toString(QUrl::FullyEncoded));
+
+		if(!route.id.isEmpty())
+			msg += QString(" route=%1").arg(QString::fromUtf8(route.id));
+
+		HttpResponseData resp = rs->responseData();
+
+		if(accepted)
+		{
+			msg += " hold";
+		}
+		else if(resp.code != -1 && !si->unclean)
+		{
+			msg += QString(" code=%1 %2").arg(QString::number(resp.code), QString::number(rs->responseBodySize()));
+		}
+		else
+		{
+			msg += " error";
+		}
+
+		if(rs->isRetry())
+			msg += " retry";
+
+		if(shared)
+			msg += QString().sprintf(" shared=%p", this);
+
+		log_info("%s", qPrintable(msg));
+	}
+
 public slots:
 	void inRequest_readyRead()
 	{
@@ -645,6 +838,8 @@ public slots:
 			responseData.headers = zhttpRequest->responseHeaders();
 			responseBody += zhttpRequest->readBody(MAX_INITIAL_BUFFER);
 
+			acceptResponseData = responseData;
+
 			total += responseBody.size();
 			log_debug("proxysession: %p recv total: %d", q, total);
 
@@ -662,19 +857,93 @@ public slots:
 				}
 				else
 				{
-					foreach(const HttpHeader &h, responseData.headers)
+					if(proxyInitialResponse && responseData.headers.get("Grip-Hold") == "stream")
 					{
-						foreach(const QByteArray &hp, acceptHeaderPrefixes)
+						// sending the initial response from the proxy means
+						//   we need to do some of the handler's job here
+
+						// NOTE: if we ever need to do more than what's
+						//   below, we should consider querying the handler
+						//   to perform these things while still letting
+						//   the proxy send the response body
+
+						// no content length
+						responseData.headers.removeAll("Content-Length");
+
+						// interpret grip-status
+						QByteArray statusHeader = responseData.headers.get("Grip-Status");
+						if(!statusHeader.isEmpty())
 						{
-							if(qstrnicmp(h.first.data(), hp.data(), hp.length()) == 0)
+							QByteArray codeStr;
+							QByteArray reason;
+
+							int at = statusHeader.indexOf(' ');
+							if(at != -1)
 							{
-								doAccept = true;
-								break;
+								codeStr = statusHeader.mid(0, at);
+								reason = statusHeader.mid(at + 1);
+							}
+							else
+							{
+								codeStr = statusHeader;
+							}
+
+							bool _ok;
+							responseData.code = codeStr.toInt(&_ok);
+							if(!_ok || responseData.code < 0 || responseData.code > 999)
+							{
+								// this may output a misleading error message
+								cannotAcceptAll();
+								return;
+							}
+
+							if(reason.isEmpty())
+								reason = StatusReasons::getReason(responseData.code);
+
+							responseData.reason = reason;
+						}
+
+						// strip any grip headers
+						for(int n = 0; n < responseData.headers.count(); ++n)
+						{
+							const HttpHeader &h = responseData.headers[n];
+
+							bool prefixed = false;
+							foreach(const QByteArray &hp, acceptHeaderPrefixes)
+							{
+								if(qstrnicmp(h.first.data(), hp.data(), hp.length()) == 0)
+								{
+									prefixed = true;
+									break;
+								}
+							}
+
+							if(prefixed)
+							{
+								responseData.headers.removeAt(n);
+								--n; // adjust position
 							}
 						}
 
-						if(doAccept)
-							break;
+						// we'll let the proxy send normally, then accept afterwards
+						acceptAfterResponding = true;
+					}
+					else
+					{
+						foreach(const HttpHeader &h, responseData.headers)
+						{
+							foreach(const QByteArray &hp, acceptHeaderPrefixes)
+							{
+								if(qstrnicmp(h.first.data(), hp.data(), hp.length()) == 0)
+								{
+									doAccept = true;
+									break;
+								}
+							}
+
+							if(doAccept)
+								break;
+						}
 					}
 				}
 			}
@@ -683,7 +952,7 @@ public slots:
 			{
 				if(!buffering)
 				{
-					rejectAll(502, "Bad Gateway", "Request too large to accept GRIP instruct.");
+					rejectAll(400, "Bad Request", "Request too large to accept GRIP instruct.");
 					return;
 				}
 
@@ -703,6 +972,7 @@ public slots:
 				foreach(SessionItem *si, sessionItems)
 				{
 					si->state = SessionItem::Responding;
+					si->startedResponse = true;
 					si->rs->startResponse(responseData.code, responseData.reason, responseData.headers);
 
 					if(!responseBody.isEmpty())
@@ -796,6 +1066,8 @@ public slots:
 		SessionItem *si = sessionItemsBySession.value(rs);
 		assert(si);
 
+		logFinished(si);
+
 		QPointer<QObject> self = this;
 		emit q->requestSessionDestroyed(si->rs, false);
 		if(!self)
@@ -851,7 +1123,7 @@ public slots:
 		{
 			assert(!acceptRequest);
 
-			responseData.body = responseBody.take();
+			acceptResponseData.body = responseBody.take();
 
 			AcceptData adata;
 
@@ -863,9 +1135,11 @@ public slots:
 				areq.rid = si->rs->rid();
 				areq.https = si->rs->isHttps();
 				areq.peerAddress = si->rs->peerAddress();
+				areq.debug = si->rs->debugEnabled();
 				areq.autoCrossOrigin = si->rs->autoCrossOrigin();
 				areq.jsonpCallback = si->rs->jsonpCallback();
 				areq.jsonpExtendedResponse = si->rs->jsonpExtendedResponse();
+				areq.responseCode = ss.responseCode;
 				areq.inSeq = ss.inSeq;
 				areq.outSeq = ss.outSeq;
 				areq.outCredits = ss.outCredits;
@@ -877,7 +1151,7 @@ public slots:
 			adata.requestData.body = requestBody.take();
 
 			adata.haveResponse = true;
-			adata.response = responseData;
+			adata.response = acceptResponseData;
 
 			if(haveInspectData)
 			{
@@ -890,9 +1164,10 @@ public slots:
 			adata.route = route.id;
 			adata.channelPrefix = route.prefix;
 			adata.useSession = route.session;
+			adata.responseSent = acceptAfterResponding;
 
 			acceptRequest = new AcceptRequest(acceptManager, this);
-			connect(acceptRequest, SIGNAL(finished()), SLOT(acceptRequest_finished()));
+			connect(acceptRequest, &AcceptRequest::finished, this, &Private::acceptRequest_finished);
 			acceptRequest->start(adata);
 		}
 	}
@@ -926,6 +1201,9 @@ public slots:
 
 			if(rdata.accepted)
 			{
+				foreach(SessionItem *si, sessionItems)
+					logFinished(si, true);
+
 				// the requests were paused, so deleting them will leave the peer sessions active
 
 				QList<RequestSession*> toDestroy;
@@ -950,31 +1228,51 @@ public slots:
 			}
 			else
 			{
-				// wake up receivers
-				foreach(SessionItem *si, sessionItems)
-				{
-					si->state = SessionItem::WaitingForResponse;
-					si->rs->resume();
-				}
-
 				if(rdata.response.code != -1)
-					respondAll(rdata.response.code, rdata.response.reason, rdata.response.headers, rdata.response.body);
+				{
+					if(acceptAfterResponding)
+					{
+						// can't respond again in this state, so just
+						//   disconnect
+						destroyAll();
+					}
+					else
+					{
+						// wake up receivers
+						foreach(SessionItem *si, sessionItems)
+						{
+							si->state = SessionItem::WaitingForResponse;
+							si->rs->resume();
+						}
+
+						respondAll(rdata.response.code, rdata.response.reason, rdata.response.headers, rdata.response.body);
+					}
+				}
 				else
+				{
 					cannotAcceptAll();
+				}
 			}
 		}
 		else
 		{
+			// wake up receivers and reject
+
+			if(acceptRequest->errorCondition() == ZrpcRequest::ErrorFormat && ((ZrpcRequest *)acceptRequest)->result().type() == QVariant::ByteArray)
+			{
+				QString errorString = QString::fromUtf8(((ZrpcRequest *)acceptRequest)->result().toByteArray());
+				QString msg = "Error while proxying to origin.";
+				QString dmsg = QString("Failed to parse accept instructions: %1").arg(errorString);
+
+				rejectAll(502, "Bad Gateway", msg, dmsg);
+			}
+			else
+			{
+				cannotAcceptAll();
+			}
+
 			delete acceptRequest;
 			acceptRequest = 0;
-
-			// wake up receivers and reject
-			foreach(SessionItem *si, sessionItems)
-			{
-				si->state = SessionItem::WaitingForResponse;
-				si->rs->resume();
-			}
-			cannotAcceptAll();
 		}
 	}
 };
@@ -1006,6 +1304,11 @@ void ProxySession::setDefaultUpstreamKey(const QByteArray &key)
 	d->defaultUpstreamKey = key;
 }
 
+void ProxySession::setAcceptXForwardedProtocol(bool enabled)
+{
+	d->acceptXForwardedProtocol = enabled;
+}
+
 void ProxySession::setUseXForwardedProtocol(bool enabled)
 {
 	d->useXForwardedProtocol = enabled;
@@ -1022,6 +1325,11 @@ void ProxySession::setOrigHeadersNeedMark(const QList<QByteArray> &names)
 	d->origHeadersNeedMark = names;
 }
 
+void ProxySession::setProxyInitialResponseEnabled(bool enabled)
+{
+	d->proxyInitialResponse = enabled;
+}
+
 void ProxySession::setInspectData(const InspectData &idata)
 {
 	d->haveInspectData = true;
