@@ -144,6 +144,7 @@ public:
 	ZhttpRequest *zhttpRequest;
 	HttpRequestData requestData;
 	DomainMap::Entry route;
+	bool debug;
 	bool autoCrossOrigin;
 	InspectRequest *inspectRequest;
 	InspectData idata;
@@ -152,6 +153,7 @@ public:
 	QByteArray jsonpCallback;
 	bool jsonpExtendedResponse;
 	HttpResponseData responseData;
+	int responseBodySize;
 	BufferList out;
 	bool responseBodyFinished;
 	bool pendingResponseUpdate;
@@ -159,6 +161,7 @@ public:
 	bool isRetry;
 	QList<QByteArray> jsonpExtractableHeaders;
 	int prefetchSize;
+	bool needPause;
 
 	Private(RequestSession *_q, DomainMap *_domainMap, SockJsManager *_sockJsManager, ZrpcManager *_inspectManager, ZrpcChecker *_inspectChecker, ZrpcManager *_acceptManager) :
 		QObject(_q),
@@ -170,14 +173,17 @@ public:
 		inspectChecker(_inspectChecker),
 		acceptManager(_acceptManager),
 		zhttpRequest(0),
+		debug(false),
 		autoCrossOrigin(false),
 		inspectRequest(0),
 		acceptRequest(0),
 		jsonpExtendedResponse(false),
+		responseBodySize(0),
 		responseBodyFinished(false),
 		pendingResponseUpdate(false),
 		isRetry(false),
-		prefetchSize(0)
+		prefetchSize(0),
+		needPause(false)
 	{
 		jsonpExtractableHeaders += "Cache-Control";
 	}
@@ -214,7 +220,7 @@ public:
 		requestData.uri = req->requestUri();
 		requestData.headers = req->requestHeaders();
 
-		log_info("IN id=%s, %s %s", rid.second.data(), qPrintable(requestData.method), requestData.uri.toEncoded().data());
+		log_debug("IN id=%s, %s %s", rid.second.data(), qPrintable(requestData.method), requestData.uri.toEncoded().data());
 
 		bool isHttps = (requestData.uri.scheme() == "https");
 		QString host = requestData.uri.host();
@@ -233,11 +239,17 @@ public:
 			return;
 		}
 
-		connect(zhttpRequest, SIGNAL(error()), SLOT(zhttpRequest_error()));
-		connect(zhttpRequest, SIGNAL(paused()), SLOT(zhttpRequest_paused()));
+		connect(zhttpRequest, &ZhttpRequest::error, this, &Private::zhttpRequest_error);
+		connect(zhttpRequest, &ZhttpRequest::paused, this, &Private::zhttpRequest_paused);
 
-		if(!route.isNull() && route.autoCrossOrigin)
-			autoCrossOrigin = true;
+		if(!route.isNull())
+		{
+			if(route.debug)
+				debug = true;
+
+			if(route.autoCrossOrigin)
+				autoCrossOrigin = true;
+		}
 
 		if(autoCrossOrigin)
 		{
@@ -294,14 +306,14 @@ public:
 
 		state = Prefetching;
 
-		connect(zhttpRequest, SIGNAL(readyRead()), SLOT(zhttpRequest_readyRead()));
+		connect(zhttpRequest, &ZhttpRequest::readyRead, this, &Private::zhttpRequest_readyRead);
 		processIncomingRequest();
 	}
 
 	void startRetry()
 	{
-		connect(zhttpRequest, SIGNAL(error()), SLOT(zhttpRequest_error()));
-		connect(zhttpRequest, SIGNAL(paused()), SLOT(zhttpRequest_paused()));
+		connect(zhttpRequest, &ZhttpRequest::error, this, &Private::zhttpRequest_error);
+		connect(zhttpRequest, &ZhttpRequest::paused, this, &Private::zhttpRequest_paused);
 
 		state = WaitingForResponse;
 
@@ -335,7 +347,7 @@ public:
 			{
 				// we've read enough body to start inspection
 
-				disconnect(zhttpRequest, SIGNAL(readyRead()), this, SLOT(zhttpRequest_readyRead()));
+				disconnect(zhttpRequest, &ZhttpRequest::readyRead, this, &Private::zhttpRequest_readyRead);
 
 				state = Inspecting;
 				requestData.body = in.toByteArray();
@@ -349,7 +361,7 @@ public:
 
 					if(inspectChecker->isInterfaceAvailable())
 					{
-						connect(inspectRequest, SIGNAL(finished()), SLOT(inspectRequest_finished()));
+						connect(inspectRequest, &InspectRequest::finished, this, &Private::inspectRequest_finished);
 						inspectChecker->watch(inspectRequest);
 						inspectRequest->start(requestData, truncated, route.session);
 					}
@@ -380,7 +392,7 @@ public:
 				//   disallow sharing before passing to proxysession. at that
 				//   point, proxysession will read the remainder of the data
 
-				disconnect(zhttpRequest, SIGNAL(readyRead()), this, SLOT(zhttpRequest_readyRead()));
+				disconnect(zhttpRequest, &ZhttpRequest::readyRead, this, &Private::zhttpRequest_readyRead);
 
 				state = WaitingForResponse;
 				requestData.body = in.take();
@@ -536,7 +548,7 @@ public:
 			callback = parsePercentEncoding(query.queryItemValue(callbackParam, QUrl::FullyEncoded).toUtf8());
 			if(callback.isEmpty())
 			{
-				log_warning("requestsession: id=%s invalid callback parameter, rejecting", rid.second.data());
+				log_debug("requestsession: id=%s invalid callback parameter, rejecting", rid.second.data());
 				*ok = false;
 				*errorMessage = "Invalid callback parameter.";
 				return false;
@@ -553,7 +565,7 @@ public:
 			method = QString::fromLatin1(parsePercentEncoding(query.queryItemValue("_method", QUrl::FullyEncoded).toUtf8()));
 			if(!validMethod(method))
 			{
-				log_warning("requestsession: id=%s invalid _method parameter, rejecting", rid.second.data());
+				log_debug("requestsession: id=%s invalid _method parameter, rejecting", rid.second.data());
 				*ok = false;
 				*errorMessage = "Invalid _method parameter.";
 				return false;
@@ -569,7 +581,7 @@ public:
 			QJsonDocument doc = QJsonDocument::fromJson(parsePercentEncoding(query.queryItemValue("_headers", QUrl::FullyEncoded).toUtf8()), &e);
 			if(e.error != QJsonParseError::NoError || !doc.isObject())
 			{
-				log_warning("requestsession: id=%s invalid _headers parameter, rejecting", rid.second.data());
+				log_debug("requestsession: id=%s invalid _headers parameter, rejecting", rid.second.data());
 				*ok = false;
 				*errorMessage = "Invalid _headers parameter.";
 				return false;
@@ -584,7 +596,7 @@ public:
 
 				if(vit.value().type() != QVariant::String)
 				{
-					log_warning("requestsession: id=%s invalid _headers parameter, rejecting", rid.second.data());
+					log_debug("requestsession: id=%s invalid _headers parameter, rejecting", rid.second.data());
 					*ok = false;
 					*errorMessage = "Invalid _headers parameter.";
 					return false;
@@ -612,7 +624,7 @@ public:
 				body = parsePercentEncoding(query.queryItemValue(bodyParam, QUrl::FullyEncoded).toUtf8());
 				if(body.isNull())
 				{
-					log_warning("requestsession: id=%s invalid body parameter, rejecting", rid.second.data());
+					log_debug("requestsession: id=%s invalid body parameter, rejecting", rid.second.data());
 					*ok = false;
 					*errorMessage = "Invalid body parameter.";
 					return false;
@@ -631,7 +643,7 @@ public:
 				body = parsePercentEncoding(query.queryItemValue("_body").toUtf8());
 				if(body.isNull())
 				{
-					log_warning("requestsession: id=%s invalid _body parameter, rejecting", rid.second.data());
+					log_debug("requestsession: id=%s invalid _body parameter, rejecting", rid.second.data());
 					*ok = false;
 					*errorMessage = "Invalid _body parameter.";
 					return false;
@@ -725,6 +737,7 @@ public slots:
 			areq.rid = rid;
 			areq.https = zhttpRequest->requestUri().scheme() == "https";
 			areq.peerAddress = zhttpRequest->peerAddress();
+			areq.debug = debug;
 			areq.autoCrossOrigin = autoCrossOrigin;
 			areq.jsonpCallback = jsonpCallback;
 			areq.jsonpExtendedResponse = jsonpExtendedResponse;
@@ -746,7 +759,7 @@ public slots:
 			adata.channelPrefix = route.prefix;
 
 			acceptRequest = new AcceptRequest(acceptManager, this);
-			connect(acceptRequest, SIGNAL(finished()), SLOT(acceptRequest_finished()));
+			connect(acceptRequest, &AcceptRequest::finished, this, &Private::acceptRequest_finished);
 			acceptRequest->start(adata);
 		}
 		else
@@ -786,7 +799,7 @@ public slots:
 
 			// successful inspect indicated we should not proxy. in that case,
 			//   collect the body and accept
-			connect(zhttpRequest, SIGNAL(readyRead()), SLOT(zhttpRequest_readyRead()));
+			connect(zhttpRequest, &ZhttpRequest::readyRead, this, &Private::zhttpRequest_readyRead);
 			processIncomingRequest();
 		}
 		else
@@ -797,7 +810,7 @@ public slots:
 				//   request body, so let's try to read it now
 				state = Receiving;
 
-				connect(zhttpRequest, SIGNAL(readyRead()), SLOT(zhttpRequest_readyRead()));
+				connect(zhttpRequest, &ZhttpRequest::readyRead, this, &Private::zhttpRequest_readyRead);
 				processIncomingRequest();
 			}
 			else
@@ -892,6 +905,7 @@ public slots:
 						headers += HttpHeader("Content-Length", QByteArray::number(body.size()));
 						zhttpRequest->beginResponse(500, "Internal Server Error", headers);
 						zhttpRequest->writeBody(body);
+						responseBodySize += body.size();
 						zhttpRequest->endBody();
 						emit q->errorResponding();
 						return;
@@ -915,7 +929,7 @@ public slots:
 						}
 					}
 
-					connect(zhttpRequest, SIGNAL(bytesWritten(int)), SLOT(zhttpRequest_bytesWritten(int)));
+					connect(zhttpRequest, &ZhttpRequest::bytesWritten, this, &Private::zhttpRequest_bytesWritten);
 
 					zhttpRequest->beginResponse(200, "OK", headers);
 
@@ -923,6 +937,7 @@ public slots:
 					jsonpTracker.specifyEncoded(buf.size(), bodyRawBuf.size());
 
 					zhttpRequest->writeBody(buf);
+					responseBodySize += buf.size();
 					zhttpRequest->endBody();
 					return;
 				}
@@ -937,6 +952,7 @@ public slots:
 					headers += HttpHeader("Content-Length", QByteArray::number(body.size()));
 					zhttpRequest->beginResponse(500, "Internal Server Error", headers);
 					zhttpRequest->writeBody(body);
+					responseBodySize += body.size();
 					zhttpRequest->endBody();
 					emit q->errorResponding();
 					return;
@@ -945,20 +961,21 @@ public slots:
 				headers += HttpHeader("Content-Type", "application/javascript");
 				headers += HttpHeader("Transfer-Encoding", "chunked");
 
-				connect(zhttpRequest, SIGNAL(bytesWritten(int)), SLOT(zhttpRequest_bytesWritten(int)));
+				connect(zhttpRequest, &ZhttpRequest::bytesWritten, this, &Private::zhttpRequest_bytesWritten);
 
 				zhttpRequest->beginResponse(200, "OK", headers);
 
 				jsonpTracker.specifyEncoded(buf.size(), 0);
 
 				zhttpRequest->writeBody(buf);
+				responseBodySize += buf.size();
 			}
 			else
 			{
 				if(autoCrossOrigin)
 					Cors::applyCorsHeaders(requestData.headers, &responseData.headers);
 
-				connect(zhttpRequest, SIGNAL(bytesWritten(int)), SLOT(zhttpRequest_bytesWritten(int)));
+				connect(zhttpRequest, &ZhttpRequest::bytesWritten, this, &Private::zhttpRequest_bytesWritten);
 
 				zhttpRequest->beginResponse(responseData.code, responseData.reason, responseData.headers);
 			}
@@ -1013,24 +1030,35 @@ public slots:
 				jsonpTracker.specifyEncoded(buf.size(), bodyRawBuf.size());
 
 				zhttpRequest->writeBody(buf);
+				responseBodySize += buf.size();
 			}
 			else
 			{
-				zhttpRequest->writeBody(out.take());
+				QByteArray buf = out.take();
+				zhttpRequest->writeBody(buf);
+				responseBodySize += buf.size();
 			}
 		}
 
 		if(responseBodyFinished)
 		{
+			assert(!needPause);
+
 			if(!jsonpCallback.isEmpty())
 			{
 				QByteArray buf = makeJsonpEnd();
 				jsonpTracker.specifyEncoded(buf.size(), 0);
 				zhttpRequest->writeBody(buf);
+				responseBodySize += buf.size();
 			}
 
 			zhttpRequest->endBody();
 		}
+		else if(needPause)
+		{
+			needPause = false;
+			zhttpRequest->pause();
+		}
 	}
 
 	void respondSuccess(const QString &message)
@@ -1081,6 +1109,21 @@ HttpRequestData RequestSession::requestData() const
 	return d->requestData;
 }
 
+HttpResponseData RequestSession::responseData() const
+{
+	return d->responseData;
+}
+
+int RequestSession::responseBodySize() const
+{
+	return d->responseBodySize;
+}
+
+bool RequestSession::debugEnabled() const
+{
+	return d->debug;
+}
+
 bool RequestSession::autoCrossOrigin() const
 {
 	return d->autoCrossOrigin;
@@ -1111,6 +1154,11 @@ ZhttpRequest *RequestSession::request()
 	return d->zhttpRequest;
 }
 
+void RequestSession::setDebugEnabled(bool enabled)
+{
+	d->debug = enabled;
+}
+
 void RequestSession::setAutoCrossOrigin(bool enabled)
 {
 	d->autoCrossOrigin = enabled;
@@ -1126,11 +1174,12 @@ void RequestSession::start(ZhttpRequest *req)
 	d->start(req);
 }
 
-void RequestSession::startRetry(ZhttpRequest *req, bool autoCrossOrigin, const QByteArray &jsonpCallback, bool jsonpExtendedResponse)
+void RequestSession::startRetry(ZhttpRequest *req, bool debug, bool autoCrossOrigin, const QByteArray &jsonpCallback, bool jsonpExtendedResponse)
 {
 	d->isRetry = true;
 	d->zhttpRequest = req;
 	d->rid = req->rid();
+	d->debug = debug;
 	d->autoCrossOrigin = autoCrossOrigin;
 	d->jsonpCallback = jsonpCallback;
 	d->jsonpExtendedResponse = jsonpExtendedResponse;
@@ -1144,9 +1193,10 @@ void RequestSession::startRetry(ZhttpRequest *req, bool autoCrossOrigin, const Q
 
 void RequestSession::pause()
 {
-	assert(d->state == Private::WaitingForResponse);
+	assert(!d->responseBodyFinished);
+	d->needPause = true;
 
-	d->zhttpRequest->pause();
+	d->responseUpdate();
 }
 
 void RequestSession::resume()
