@@ -21,6 +21,7 @@
 
 #include <assert.h>
 #include <QTimer>
+#include <QDateTime>
 #include <QUrl>
 #include <QJsonDocument>
 #include <QJsonObject>
@@ -38,6 +39,7 @@
 #include "statsmanager.h"
 #include "inspectdata.h"
 #include "connectionmanager.h"
+#include "testwebsocket.h"
 
 #define ACTIVITY_TIMEOUT 60000
 
@@ -233,6 +235,7 @@ public:
 	QByteArray defaultSigKey;
 	QByteArray defaultUpstreamKey;
 	bool passToUpstream;
+	bool acceptXForwardedProtocol;
 	bool useXForwardedProtocol;
 	XffRule xffRule;
 	XffRule xffTrustedRule;
@@ -243,15 +246,18 @@ public:
 	int inPendingBytes;
 	int outPendingBytes;
 	int outReadInProgress; // frame type or -1
+	QByteArray pathBeg;
 	QByteArray routeId;
 	QByteArray channelPrefix;
 	QList<DomainMap::Target> targets;
+	DomainMap::Target target;
 	bool acceptGripMessages;
 	QByteArray messagePrefix;
 	bool detached;
 	QString subChannel;
-	QTimer *activityTimer;
+	QDateTime activityTime;
 	QByteArray publicCid;
+	QTimer *keepAliveTimer;
 
 	Private(WsProxySession *_q, ZRoutes *_zroutes, ConnectionManager *_connectionManager, StatsManager *_statsManager, WsControlManager *_wsControlManager) :
 		QObject(_q),
@@ -264,6 +270,7 @@ public:
 		wsControlManager(_wsControlManager),
 		wsControl(0),
 		passToUpstream(false),
+		acceptXForwardedProtocol(false),
 		useXForwardedProtocol(false),
 		inSock(0),
 		outSock(0),
@@ -271,11 +278,9 @@ public:
 		outPendingBytes(0),
 		outReadInProgress(-1),
 		acceptGripMessages(false),
-		detached(false)
+		detached(false),
+		keepAliveTimer(0)
 	{
-		activityTimer = new QTimer(this);
-		connect(activityTimer, SIGNAL(timeout()), SLOT(activity_timeout()));
-		activityTimer->setSingleShot(true);
 	}
 
 	~Private()
@@ -285,6 +290,8 @@ public:
 
 	void cleanup()
 	{
+		cleanupKeepAliveTimer();
+
 		cleanupInSock();
 
 		delete outSock;
@@ -293,14 +300,6 @@ public:
 		delete wsControl;
 		wsControl = 0;
 
-		if(activityTimer)
-		{
-			activityTimer->setParent(0);
-			activityTimer->disconnect(this);
-			activityTimer->deleteLater();
-			activityTimer = 0;
-		}
-
 		if(zhttpManager)
 		{
 			zroutes->removeRef(zhttpManager);
@@ -318,6 +317,17 @@ public:
 		}
 	}
 
+	void cleanupKeepAliveTimer()
+	{
+		if(keepAliveTimer)
+		{
+			keepAliveTimer->disconnect(this);
+			keepAliveTimer->setParent(0);
+			keepAliveTimer->deleteLater();
+			keepAliveTimer = 0;
+		}
+	}
+
 	void start(WebSocket *sock, const QByteArray &_publicCid, const DomainMap::Entry &entry)
 	{
 		assert(!inSock);
@@ -327,15 +337,15 @@ public:
 		publicCid = _publicCid;
 
 		if(statsManager)
-			activityTimer->start(ACTIVITY_TIMEOUT);
+			activityTime = QDateTime::currentDateTimeUtc();
 
 		inSock = sock;
 		inSock->setParent(this);
-		connect(inSock, SIGNAL(readyRead()), SLOT(in_readyRead()));
-		connect(inSock, SIGNAL(framesWritten(int, int)), SLOT(in_framesWritten(int, int)));
-		connect(inSock, SIGNAL(peerClosed()), SLOT(in_peerClosed()));
-		connect(inSock, SIGNAL(closed()), SLOT(in_closed()));
-		connect(inSock, SIGNAL(error()), SLOT(in_error()));
+		connect(inSock, &WebSocket::readyRead, this, &Private::in_readyRead);
+		connect(inSock, &WebSocket::framesWritten, this, &Private::in_framesWritten);
+		connect(inSock, &WebSocket::peerClosed, this, &Private::in_peerClosed);
+		connect(inSock, &WebSocket::closed, this, &Private::in_closed);
+		connect(inSock, &WebSocket::error, this, &Private::in_error);
 
 		requestData.uri = inSock->requestUri();
 		requestData.headers = inSock->requestHeaders();
@@ -345,7 +355,7 @@ public:
 		if(entry.isNull())
 		{
 			log_warning("wsproxysession: %p %s has 0 routes", q, qPrintable(host));
-			reject(502, "Bad Gateway", QString("No route for host: %1").arg(host));
+			reject(false, 502, "Bad Gateway", QString("No route for host: %1").arg(host));
 			return;
 		}
 
@@ -375,13 +385,14 @@ public:
 			sigKey = defaultSigKey;
 		}
 
+		pathBeg = entry.pathBeg;
 		routeId = entry.id;
 		channelPrefix = entry.prefix;
 		targets = entry.targets;
 
 		log_debug("wsproxysession: %p %s has %d routes", q, qPrintable(host), targets.count());
 
-		bool trustedClient = ProxyUtil::manipulateRequestHeaders("wsproxysession", q, &requestData, defaultUpstreamKey, entry, sigIss, sigKey, useXForwardedProtocol, xffTrustedRule, xffRule, origHeadersNeedMark, inSock->peerAddress(), InspectData());
+		bool trustedClient = ProxyUtil::manipulateRequestHeaders("wsproxysession", q, &requestData, defaultUpstreamKey, entry, sigIss, sigKey, acceptXForwardedProtocol, useXForwardedProtocol, xffTrustedRule, xffRule, origHeadersNeedMark, inSock->peerAddress(), InspectData());
 
 		// don't proxy extensions, as we may not know how to handle them
 		requestData.headers.removeAll("Sec-WebSocket-Extensions");
@@ -399,11 +410,11 @@ public:
 	{
 		if(targets.isEmpty())
 		{
-			reject(502, "Bad Gateway", "Error while proxying to origin.");
+			reject(false, 502, "Bad Gateway", "Error while proxying to origin.");
 			return;
 		}
 
-		DomainMap::Target target = targets.takeFirst();
+		target = targets.takeFirst();
 
 		QUrl uri = requestData.uri;
 		if(target.ssl)
@@ -417,46 +428,67 @@ public:
 		subChannel = target.subChannel;
 
 		if(zhttpManager)
+		{
 			zroutes->removeRef(zhttpManager);
-
-		if(target.type == DomainMap::Target::Custom)
-		{
-			zhttpManager = zroutes->managerForRoute(target.zhttpRoute);
-			log_debug("wsproxysession: %p forwarding to %s", q, qPrintable(target.zhttpRoute.baseSpec));
-		}
-		else // Default
-		{
-			zhttpManager = zroutes->defaultManager();
-			log_debug("wsproxysession: %p forwarding to %s:%d", q, qPrintable(target.connectHost), target.connectPort);
+			zhttpManager = 0;
 		}
 
-		zroutes->addRef(zhttpManager);
-
-		if(target.overHttp)
+		if(target.type == DomainMap::Target::Test)
 		{
-			WebSocketOverHttp *woh = new WebSocketOverHttp(zhttpManager, this);
-			woh->setConnectionId(publicCid);
-			outSock = woh;
+			// for test route, auto-adjust path
+			if(!pathBeg.isEmpty())
+			{
+				int pathRemove = pathBeg.length();
+				if(pathBeg.endsWith('/'))
+					--pathRemove;
+
+				if(pathRemove > 0)
+					uri.setPath(uri.path(QUrl::FullyEncoded).mid(pathRemove));
+			}
+
+			outSock = new TestWebSocket(this);
 		}
 		else
 		{
-			// websockets don't work with zhttp req mode
-			if(zhttpManager->clientUsesReq())
+			if(target.type == DomainMap::Target::Custom)
 			{
-				reject(502, "Bad Gateway", "Error while proxying to origin.");
-				return;
+				zhttpManager = zroutes->managerForRoute(target.zhttpRoute);
+				log_debug("wsproxysession: %p forwarding to %s", q, qPrintable(target.zhttpRoute.baseSpec));
+			}
+			else // Default
+			{
+				zhttpManager = zroutes->defaultManager();
+				log_debug("wsproxysession: %p forwarding to %s:%d", q, qPrintable(target.connectHost), target.connectPort);
 			}
 
-			outSock = zhttpManager->createSocket();
-			outSock->setParent(this);
+			zroutes->addRef(zhttpManager);
+
+			if(target.overHttp)
+			{
+				WebSocketOverHttp *woh = new WebSocketOverHttp(zhttpManager, this);
+				woh->setConnectionId(publicCid);
+				outSock = woh;
+			}
+			else
+			{
+				// websockets don't work with zhttp req mode
+				if(zhttpManager->clientUsesReq())
+				{
+					reject(false, 502, "Bad Gateway", "Error while proxying to origin.");
+					return;
+				}
+
+				outSock = zhttpManager->createSocket();
+				outSock->setParent(this);
+			}
 		}
 
-		connect(outSock, SIGNAL(connected()), SLOT(out_connected()));
-		connect(outSock, SIGNAL(readyRead()), SLOT(out_readyRead()));
-		connect(outSock, SIGNAL(framesWritten(int, int)), SLOT(out_framesWritten(int, int)));
-		connect(outSock, SIGNAL(peerClosed()), SLOT(out_peerClosed()));
-		connect(outSock, SIGNAL(closed()), SLOT(out_closed()));
-		connect(outSock, SIGNAL(error()), SLOT(out_error()));
+		connect(outSock, &WebSocket::connected, this, &Private::out_connected);
+		connect(outSock, &WebSocket::readyRead, this, &Private::out_readyRead);
+		connect(outSock, &WebSocket::framesWritten, this, &Private::out_framesWritten);
+		connect(outSock, &WebSocket::peerClosed, this, &Private::out_peerClosed);
+		connect(outSock, &WebSocket::closed, this, &Private::out_closed);
+		connect(outSock, &WebSocket::error, this, &Private::out_error);
 
 		if(target.trusted)
 			outSock->setIgnorePolicies(true);
@@ -473,17 +505,19 @@ public:
 		outSock->start(uri, requestData.headers);
 	}
 
-	void reject(int code, const QByteArray &reason, const HttpHeaders &headers, const QByteArray &body)
+	void reject(bool proxied, int code, const QByteArray &reason, const HttpHeaders &headers, const QByteArray &body)
 	{
 		assert(state == Connecting);
 
 		state = Closing;
 		inSock->respondError(code, reason, headers, body);
+
+		logConnection(proxied, code, body.size());
 	}
 
-	void reject(int code, const QString &reason, const QString &errorMessage)
+	void reject(bool proxied, int code, const QString &reason, const QString &errorMessage)
 	{
-		reject(code, reason.toUtf8(), HttpHeaders(), (errorMessage + '\n').toUtf8());
+		reject(proxied, code, reason.toUtf8(), HttpHeaders(), (errorMessage + '\n').toUtf8());
 	}
 
 	void tryReadIn()
@@ -537,6 +571,8 @@ public:
 						f.data = f.data.mid(messagePrefix.size());
 						inSock->writeFrame(f);
 						inPendingBytes += f.data.size();
+
+						restartKeepAlive();
 					}
 					else if(f.type == WebSocket::Frame::Continuation)
 					{
@@ -544,12 +580,16 @@ public:
 
 						inSock->writeFrame(f);
 						inPendingBytes += f.data.size();
+
+						restartKeepAlive();
 					}
 				}
 				else
 				{
 					inSock->writeFrame(f);
 					inPendingBytes += f.data.size();
+
+					restartKeepAlive();
 				}
 
 				if(!f.more)
@@ -560,6 +600,8 @@ public:
 				// always relay non-content frames
 				inSock->writeFrame(f);
 				inPendingBytes += f.data.size();
+
+				restartKeepAlive();
 			}
 		}
 	}
@@ -575,14 +617,59 @@ public:
 
 	void tryLogActivity()
 	{
-		if(statsManager && !activityTimer->isActive())
+		if(statsManager && !activityTime.isNull())
 		{
-			statsManager->addActivity(routeId);
+			QDateTime now = QDateTime::currentDateTimeUtc();
+			if(now >= activityTime.addMSecs(ACTIVITY_TIMEOUT))
+			{
+				statsManager->addActivity(routeId);
 
-			activityTimer->start(ACTIVITY_TIMEOUT);
+				activityTime = activityTime.addMSecs((activityTime.msecsTo(now) / ACTIVITY_TIMEOUT) * ACTIVITY_TIMEOUT);
+			}
 		}
 	}
 
+	void logConnection(bool proxied, int responseCode, int responseBodySize)
+	{
+		QString msg = QString("GET %1").arg(inSock->requestUri().toString(QUrl::FullyEncoded));
+
+		if(proxied)
+		{
+			QString targetStr;
+			if(target.type == DomainMap::Target::Test)
+			{
+				targetStr = "test";
+			}
+			else if(target.type == DomainMap::Target::Custom)
+			{
+				targetStr = (target.zhttpRoute.req ? "zhttpreq/" : "zhttp/") + target.zhttpRoute.baseSpec;
+			}
+			else // Default
+			{
+				targetStr = target.connectHost + ':' + QString::number(target.connectPort);
+			}
+
+			msg += QString(" -> %2").arg(targetStr);
+			if(target.overHttp)
+				msg += "[http]";
+		}
+
+		QUrl ref = QUrl(QString::fromUtf8(inSock->requestHeaders().get("Referer")));
+		if(!ref.isEmpty())
+			msg += QString(" ref=%1").arg(ref.toString(QUrl::FullyEncoded));
+		if(!routeId.isEmpty())
+			msg += QString(" route=%1").arg(QString::fromUtf8(routeId));
+		msg += QString(" code=%1 %2").arg(QString::number(responseCode), QString::number(responseBodySize));
+
+		log_info("%s", qPrintable(msg));
+	}
+
+	void restartKeepAlive()
+	{
+		if(keepAliveTimer)
+			keepAliveTimer->start();
+	}
+
 private slots:
 	void in_readyRead()
 	{
@@ -685,10 +772,12 @@ private slots:
 			if(wsControlManager)
 			{
 				wsControl = wsControlManager->createSession(publicCid);
-				connect(wsControl, SIGNAL(sendEventReceived(const QByteArray &, const QByteArray &)), SLOT(wsControl_sendEventReceived(const QByteArray &, const QByteArray &)));
-				connect(wsControl, SIGNAL(detachEventReceived()), SLOT(wsControl_detachEventReceived()));
-				connect(wsControl, SIGNAL(cancelEventReceived()), SLOT(wsControl_cancelEventReceived()));
-				wsControl->start(channelPrefix);
+				connect(wsControl, &WsControlSession::sendEventReceived, this, &Private::wsControl_sendEventReceived);
+				connect(wsControl, &WsControlSession::keepAliveSetupEventReceived, this, &Private::wsControl_keepAliveSetupEventReceived);
+				connect(wsControl, &WsControlSession::closeEventReceived, this, &Private::wsControl_closeEventReceived);
+				connect(wsControl, &WsControlSession::detachEventReceived, this, &Private::wsControl_detachEventReceived);
+				connect(wsControl, &WsControlSession::cancelEventReceived, this, &Private::wsControl_cancelEventReceived);
+				wsControl->start(routeId, channelPrefix, inSock->requestUri());
 
 				if(!subChannel.isEmpty())
 				{
@@ -706,6 +795,8 @@ private slots:
 
 		inSock->respondSuccess(outSock->responseReason(), headers);
 
+		logConnection(true, 101, 0);
+
 		// send any pending frames
 		tryReadIn();
 	}
@@ -769,10 +860,10 @@ private slots:
 					tryAgain = true;
 					break;
 				case WebSocket::ErrorRejected:
-					reject(outSock->responseCode(), outSock->responseReason(), outSock->responseHeaders(), outSock->responseBody());
+					reject(true, outSock->responseCode(), outSock->responseReason(), outSock->responseHeaders(), outSock->responseBody());
 					break;
 				default:
-					reject(502, "Bad Gateway", "Error while proxying to origin.");
+					reject(true, 502, "Bad Gateway", "Error while proxying to origin.");
 					break;
 			}
 
@@ -793,18 +884,50 @@ private slots:
 		}
 	}
 
-	void wsControl_sendEventReceived(const QByteArray &contentType, const QByteArray &message)
+	void wsControl_sendEventReceived(WebSocket::Frame::Type type, const QByteArray &message)
 	{
+		// this method accepts a full message, which must be typed
+		if(type == WebSocket::Frame::Continuation)
+			return;
+
 		// only send if we can, otherwise drop
 		if(inSock && inSock->canWrite())
 		{
-			if(contentType == "binary")
-				inSock->writeFrame(WebSocket::Frame(WebSocket::Frame::Binary, message, false));
-			else
-				inSock->writeFrame(WebSocket::Frame(WebSocket::Frame::Text, message, false));
+			inSock->writeFrame(WebSocket::Frame(type, message, false));
 
 			inPendingBytes += message.size();
 		}
+
+		restartKeepAlive();
+	}
+
+	void wsControl_keepAliveSetupEventReceived(bool enable, int timeout)
+	{
+		if(enable)
+		{
+			if(!keepAliveTimer)
+			{
+				keepAliveTimer = new QTimer(this);
+				connect(keepAliveTimer, &QTimer::timeout, this, &Private::keepAliveTimer_timeout);
+				keepAliveTimer->setSingleShot(true);
+			}
+
+			keepAliveTimer->setInterval(timeout * 1000);
+			keepAliveTimer->start();
+		}
+		else
+		{
+			cleanupKeepAliveTimer();
+		}
+	}
+
+	void wsControl_closeEventReceived(int code)
+	{
+		if(!detached && outSock && outSock->state() != WebSocket::Closing)
+			outSock->close();
+
+		if(inSock && inSock->state() != WebSocket::Closing)
+			inSock->close(code);
 	}
 
 	void wsControl_detachEventReceived()
@@ -832,9 +955,9 @@ private slots:
 		tryFinish();
 	}
 
-	void activity_timeout()
+	void keepAliveTimer_timeout()
 	{
-		// nothing to do
+		wsControl->sendNeedKeepAlive();
 	}
 };
 
@@ -870,6 +993,11 @@ void WsProxySession::setDefaultUpstreamKey(const QByteArray &key)
 	d->defaultUpstreamKey = key;
 }
 
+void WsProxySession::setAcceptXForwardedProtocol(bool enabled)
+{
+	d->acceptXForwardedProtocol = enabled;
+}
+
 void WsProxySession::setUseXForwardedProtocol(bool enabled)
 {
 	d->useXForwardedProtocol = enabled;
